from tensorflow.keras.layers import Input
from tensorflow.keras import layers
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import ZeroPadding2D

from tensorflow.keras.models import Model

import tensorflow.keras.backend as K

from dl.machine_learning.dextr.classifiers import build_pyramid_pooling_module, BN


def identity_block(input_tensor, kernel_size, filters, stage, block, dilation=1):
    """The identity block is the block that has no conv layer at shortcut.

    # Arguments
        input_tensor: input tensor
        kernel_size: defualt 3, the kernel size of middle conv layer at main path
        filters: list of integers, the filterss of 3 conv layer at main path
        stage: integer, current stage label, used for generating layer names
        block: 'a','b'..., current block label, used for generating layer names
        dilation: dilation of the intermediate convolution

    # Returns
        Output tensor for the block.
    """
    filters1, filters2, filters3 = filters
    if K.image_data_format() == 'channels_last':
        bn_axis = 3
    else:
        bn_axis = 1
    conv_name_base = 'res' + str(stage) + block + '_branch'
    bn_name_base = 'bn' + str(stage) + block + '_branch'

    x = Conv2D(filters1, (1, 1), name=conv_name_base + '2a', use_bias=False)(input_tensor)
    x = BN(axis=bn_axis, name=bn_name_base + '2a')(x)
    x = Activation('relu')(x)

    x = Conv2D(filters2, kernel_size, padding='same', name=conv_name_base + '2b', use_bias=False, dilation_rate=dilation)(x)
    x = BN(axis=bn_axis, name=bn_name_base + '2b')(x)
    x = Activation('relu')(x)

    x = Conv2D(filters3, (1, 1), name=conv_name_base + '2c', use_bias=False)(x)
    x = BN(axis=bn_axis, name=bn_name_base + '2c')(x)

    x = layers.add([x, input_tensor])
    x = Activation('relu')(x)
    return x


def conv_block(input_tensor, kernel_size, filters, stage, block, strides=(1, 1), dilation=1):
    """conv_block is the block that has a conv layer at shortcut

    # Arguments
        input_tensor: input tensor
        kernel_size: defualt 3, the kernel size of middle conv layer at main path
        filters: list of integers, the filterss of 3 conv layer at main path
        stage: integer, current stage label, used for generating layer names
        block: 'a','b'..., current block label, used for generating layer names

    # Returns
        Output tensor for the block.

    Note that from stage 3, the first conv layer at main path is with strides=(2,2)
    And the shortcut should have strides=(2,2) as well
    """
    filters1, filters2, filters3 = filters
    if K.image_data_format() == 'channels_last':
        bn_axis = 3
    else:
        bn_axis = 1
    conv_name_base = 'res' + str(stage) + block + '_branch'
    bn_name_base = 'bn' + str(stage) + block + '_branch'

    x = Conv2D(filters1, (1, 1), strides=strides, name=conv_name_base + '2a', use_bias=False)(input_tensor)
    x = BN(axis=bn_axis, name=bn_name_base + '2a')(x)
    x = Activation('relu')(x)

    x = Conv2D(filters2, kernel_size, padding='same', name=conv_name_base + '2b', use_bias=False, dilation_rate=dilation)(x)
    x = BN(axis=bn_axis, name=bn_name_base + '2b')(x)
    x = Activation('relu')(x)

    x = Conv2D(filters3, (1, 1), name=conv_name_base + '2c', use_bias=False)(x)
    x = BN(axis=bn_axis, name=bn_name_base + '2c')(x)

    shortcut = Conv2D(filters3, (1, 1), strides=strides, name=conv_name_base + '1', use_bias=False)(input_tensor)
    shortcut = BN(axis=bn_axis, name=bn_name_base + '1')(shortcut)

    x = layers.add([x, shortcut])
    x = Activation('relu')(x)
    return x


def ResNet101(input_tensor=None):

    img_input = input_tensor
    if K.image_data_format() == 'channels_last':
        bn_axis = 3
    else:
        bn_axis = 1

    x = ZeroPadding2D((3, 3))(img_input)
    x = Conv2D(64, (7, 7), strides=(2, 2), name='conv1', use_bias=False)(x)
    x = BN(axis=bn_axis, name='bn_conv1')(x)
    x = Activation('relu')(x)
    x = ZeroPadding2D((1, 1))(x)
    x = MaxPooling2D((3, 3), strides=(2, 2), padding='valid')(x)

    x = conv_block(x, 3, [64, 64, 256], stage=2, block='a')
    x = identity_block(x, 3, [64, 64, 256], stage=2, block='b')
    x = identity_block(x, 3, [64, 64, 256], stage=2, block='c')

    x = conv_block(x, 3, [128, 128, 512], stage=3, block='a', strides=(2, 2))
    x = identity_block(x, 3, [128, 128, 512], stage=3, block='b')
    x = identity_block(x, 3, [128, 128, 512], stage=3, block='c')
    x = identity_block(x, 3, [128, 128, 512], stage=3, block='d')

    x = conv_block(x, 3, [256, 256, 1024], stage=4, block='a', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='b', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='c', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='d', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='e', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='f', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='g', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='h', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='i', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='j', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='k', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='l', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='m', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='n', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='o', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='p', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='q', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='r', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='s', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='t', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='u', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='v', dilation=2)
    x = identity_block(x, 3, [256, 256, 1024], stage=4, block='w', dilation=2)

    x = conv_block(x, 3, [512, 512, 2048], stage=5, block='a', dilation=4)
    x = identity_block(x, 3, [512, 512, 2048], stage=5, block='b', dilation=4)
    x = identity_block(x, 3, [512, 512, 2048], stage=5, block='c', dilation=4)

    return x


def build_network(nb_classes, input_shape, resnet_layers=101, classifier='psp', sigmoid=False, output_size=None,
                  num_input_channels=4):
    """Build Network"""
    inp = Input((input_shape[0], input_shape[1], num_input_channels))
    if resnet_layers == 101:
        res = ResNet101(inp)
    else:
        ValueError('Resnet {} does not exist'.format(resnet_layers))
    if classifier == 'psp':
        x = build_pyramid_pooling_module(res, input_shape, nb_classes, sigmoid=sigmoid, output_size=output_size)
    else:
        raise ValueError('Classifier not implemented.')
    model = Model(inputs=inp, outputs=x)

    return model




import cv2
from PIL import Image
import numpy as np

def supportedImageFormats():
    return ['.png', '.jpg', '.jpeg', '.bmp', '.tif', '.tiff']

class ImageFile():
    def __init__(self, path, asBGR = False):
        self._image = None
        self._stack = False
        self.readImage(path)
        self.path = path
        self.brightness = 0
        self.contrast = 1
        if asBGR:
            self.normalizeImage(force8bit = True)
            self.convertToBGR()

    def readImage(self, path):
        try: # open stack
            image = Image.open(path)
            image.seek(0)
            
            
            ### conversion to array hangs up for some images
            ### see https://github.com/python-pillow/Pillow/issues/5749
            # supportedmodes = ['1', 'L', 'P', 'RGB', 'RGBA', 'CMYK', 'YCbCr', 'LAB', 'HSV', 'I', 'F']

            # mostly 16 bit images and lzw compressed files
            # if image.mode not in supportedmodes or image.info["compression"] == 'tiff_lzw': 
            if image.info["compression"] == 'tiff_lzw': 
                raise
            
            frames = image.n_frames
            images = []
            for i in range(frames):
                image.seek(i)
                images.append(np.array(image))

            self._image = np.array(images)
            # convert to bgr if color
            if len(self._image.shape)>=4: 
                self._image = self._image[:, :, :, [2, 1, 0],...].copy()
            
            if frames == 1:
                self._stack = False
                self._image = np.squeeze(self._image)
            elif frames > 1:
                self._stack = True  
                
        except:
            self._image = cv2.imread(path, cv2.IMREAD_UNCHANGED)
            self._stack = False
            if self._image is None:
                raise
            
        
    def convertToBGR(self):
        if self._image is None:
            return
        if self._stack:
            if len(self._image.shape) == 3 or self._image.shape[3] == 1:
                images = []
                for i in range (self._image.shape[0]):
                    image = cv2.cvtColor(self._image[i], cv2.COLOR_GRAY2BGR )
                    images.append(image)
                self._image = np.array(images) 
            if len(self._image.shape) == 5:
                images = []
                for i in range (self._image.shape[0]):
                    image = cv2.cvtColor(self._image[i], cv2.CV_BGRA2BGR )
                    images.append(image)
                self._image = np.array(images) 
            return
        else:
            if len(self._image.shape) == 2 or self._image.shape[2] == 1:
                self._image = cv2.cvtColor(self._image, cv2.COLOR_GRAY2BGR )
            if self._image.shape[2] == 4:
                self._image = cv2.cvtColor(self._image, cv2.COLOR_BGRA2BGR )
 
    def normalizeImage(self, force8bit = False):
        if self._image is None:
            return

        min_ = np.min(self._image)
        max_ = np.max(self._image)
        if self._image.dtype == np.uint16 and not force8bit:
            self._image = (self._image - min_)/(max_-min_) * 65536
            self._image = self._image.astype('uint16')
        else:
            self._image = (self._image - min_)/(max_-min_) * 255
            self._image = self._image.astype('uint8')
            
    def width(self):
        ch = 2 if self._stack else 1
        return self._image.shape[ch]
        
    def height(self):
        ch = 1 if self._stack else 0
        return self._image.shape[ch]
            
    def isStack(self):
        return True if self._stack else False
    
    def numOfImagesInStack(self):
        if not self._stack:
            return 1
        else:
            return self._image.shape[0]

    def convert2DeepLearningInput(self, monochrome, image):
        if len(image.shape) == 4:
            image = cv2.cvtColor(image, cv2.COLOR_BGRA2BGR )
        if len(image.shape) == 2:
            if not monochrome:
                image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR )
            else:
                image = image[..., np.newaxis]
        elif len(image.shape) == 3:
            if monochrome and image.shape[2] > 1:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY )
                # atm I dont know if bgr2gray squeezes
                if len(image.shape) == 2:
                    image = image[..., np.newaxis]    
        return image
   
    def adjustBrightnessContrast(self, image):
        if self.contrast == 1 and self.brightness == 0:
            return image
        # in general this can be done more efficient and without the memory use of a 16 bit image
        im = (image.astype(np.int16) * self.contrast) + self.brightness
        im[im<0] = 0
        im[im>255]=255
        return im.astype(np.uint8)

    def getDLInputImage(self,monochrome,frame=0):
        return self.convert2DeepLearningInput(monochrome, self.getImage(frame))

    def getCorrectedImage(self, frame = 0):
        return self.adjustBrightnessContrast(self.getImage(frame))

    def getImage(self, frame = 0):
        if self._stack:
            return self._image[frame]
        else:
            return self._image


def read_tiff(path):
    # returns a stack of [num, w,h,(ch)]
    img = Image.open(path)
    images = []
    for i in range(img.n_frames):
        img.seek(i)
        images.append(np.array(img))
    return np.array(images)

def normalizeImage(image):
    norm_image = np.zeros(image.shape)
    norm_image = cv2.normalize(image,norm_image, 0, 255, cv2.NORM_MINMAX)
    image = norm_image.astype('uint8')
    return image

def convertToBGR(image):
    if len(image.shape) == 2 or image.shape[2] == 1:
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR )
    if len(image.shape) == 4:
        image = cv2.cvtColor(image, cv2.CV_BGRA2BGR )
    return image

def readImageAsBGR(path):     
    # always returns 3-channel normalized opencv image with 8-bit depth -> for displaying purposes mainly
    image = cv2.imread(path, cv2.IMREAD_UNCHANGED)
    image = normalizeImage(image)
    image = convertToBGR(image)
    return image

def readNumOfImageFrames(path):
    try:
        image = Image.open(path)
        image.seek(0)
        return image.n_frames
    except:
        return 1



# -*- coding: utf-8 -*-
"""
Created on Mon Jul 01 13:18:35 2020

@author: Koerber
"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.painter.Painter import Painter

import os
from ui.Tools import canvasTool
from utils.shapes.ImageLabel import ImageLabel, saveImageLabel, loadImageLabel

class ImageLabelPainter(Painter):
    def __init__(self, canvas):
        super(ImageLabelPainter,self).__init__(canvas)
        self.tools.append(canvasTool.drag)
        self.tools.append(canvasTool.setimageclass)
        self.tools.append(canvasTool.assignimageclass)
        self.imagelabel = None

    def enableDrawBackgroundMode(self):
        super(ImageLabelPainter, self).enableDrawBackgroundMode()
        self.tools.append(canvasTool.setimageclass)
        self.tools.append(canvasTool.assignimageclass)

    def disableDrawBackgroundMode(self):
        super(ImageLabelPainter, self).disableDrawBackgroundMode()
        self.tools.append(canvasTool.setimageclass)
        self.tools.append(canvasTool.assignimageclass)

    def draw(self):
        super(ImageLabelPainter, self).draw()
        self.drawLabel()
        
    def clearFOV(self):
        # do nothing
        pass
    
    def addShapes(self,shapes):
        self.addImageLabel(shapes.classlabel)
    
    def clear(self):
        super(ImageLabelPainter, self).clear()
        self.imagelabel = None

    def addImageLabel(self,classlabel=None):
        if classlabel == None:
            classlabel = self.canvas.parent.activeClass()
        self.imagelabel = ImageLabel(classlabel)
        self.canvas.redrawImage()

    def load(self):
        self.clear()
        if self.canvas.parent.files.CurrentLabelPath() is not None and os.path.exists(self.canvas.parent.files.CurrentLabelPath()):    
            self.imagelabel, background = loadImageLabel(self.canvas.parent.files.CurrentLabelPath())
            if background:
                self.contours.addShapes(background)

    def save(self):
        if len(self.trackChanges) > self.maxChanges:
            self.trackChanges.pop(0)
        self.trackChanges.append((self.imagelabel, self.backgroundshapes.copy()))
        self.canvas.parent.checkIfUndoPossible()

        if self.imagelabel or not self.contours.empty():
            self.canvas.parent.ensureLabelFolder()
            saveImageLabel(self.imagelabel, self.canvas.parent.files.CurrentLabelPath(), background = self.backgroundshapes)
        else:
            self.canvas.parent.deleteLabel()

    def undo(self):
        if len(self.trackChanges) < 2:
            return
        self.clear()
        self.imagelabel, background = self.trackChanges[-2]
        self.trackChanges.pop()
        self.trackChanges.pop()
        if background:
            self.contours.addShapes(background)
        self.canvas.parent.checkIfUndoPossible()
        
    def checkForChanges(self):
        # to do
        self.canvas.redrawImage()
        
    def drawLabel(self):
        if not self.imagelabel:
            return
        painter = self.getPainter()
        color = self.canvas.parent.ClassColor(self.imagelabel.classlabel)
        self.setPainterColor(painter, color)
        w = self.canvas.image().width()
        h = self.canvas.image().height()
        p1 = QPoint(0,0)
        p2 = QPoint(w-1,0)
        p3 = QPoint(w-1,h-1)
        p4 = QPoint (0,h-1)
        painter.drawLine(p1, p2)
        painter.drawLine(p2, p3)
        painter.drawLine(p3, p4)
        painter.drawLine(p4, p1)





# adapted from https://github.com/scaelles/DEXTR-KerasTensorflow
# licence GPLv3, https://github.com/scaelles/DEXTR-KerasTensorflow/blob/master/LICENSE

import numpy as np
from dl.machine_learning.dextr.dextr import DEXTR
import dl.machine_learning.dextr.helpers as helpers
from dl.utils.dl_downloads import get_dextr

class DEXTR_Segmentation():
    def __init__(self):
        self.pad = 50
        self.thres = 0.8
        self.modelName = None 
        self.net = None
        
    def loadModel(self):
        try:
            if self.net is None:
                if self.modelName is None:
                    self.modelName =  get_dextr()
                self.net = DEXTR(nb_classes=1, resnet_layers=101, input_shape=(512, 512), weights=self.modelName, num_input_channels=4, classifier='psp', sigmoid=True)
        except:
            # can not load file
            self.net = None

    def predict(self, image, extreme_points_ori):
        if not self.net:
            self.loadModel()
            if not self.net:
                return 

        #  Crop image to the bounding box from the extreme points and resize
        bbox = helpers.get_bbox(image, points=extreme_points_ori, pad=self.pad, zero_pad=True)
        crop_image = helpers.crop_from_bbox(image, bbox, zero_pad=True)
        resize_image = helpers.fixed_resize(crop_image, (512, 512)).astype(np.float32)

        #  Generate extreme point heat map normalized to image values
        extreme_points = extreme_points_ori - [np.min(extreme_points_ori[:, 0]), np.min(extreme_points_ori[:, 1])] + [self.pad, self.pad]
        extreme_points = (512 * extreme_points * [1 / crop_image.shape[1], 1 / crop_image.shape[0]]).astype(np.int)
        extreme_heatmap = helpers.make_gt(resize_image, extreme_points, sigma=10)
        extreme_heatmap = helpers.cstm_normalize(extreme_heatmap, 255)

        #  Concatenate inputs and convert to tensor
        input_dextr = np.concatenate((resize_image, extreme_heatmap[:, :, np.newaxis]), axis=2)

        # Run a forward pass
        pred = self.net.model.predict(input_dextr[np.newaxis, ...])[0, :, :, 0]
        result = helpers.crop2fullmask(pred, bbox, im_size=image.shape[:2], zero_pad=True, relax=self.pad) > self.thres
        return result.astype(np.int)


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 20 13:42:08 2020

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import csv
import os
import glob
import concurrent.futures
from utils.shapes.Contour import loadContours
from utils.shapes.Point import loadPoints
from ui.Tools import canvasTool
from ui.style import styleForm
from ui.ui_utils import DCDButton
from utils.Image import ImageFile
from dl.method.mode import dlMode
import cv2


class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Object Detection Settings') 

        styleForm(Form)
        self.centralWidget = QWidget(Form)

        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.vlayout.setSpacing(6)
        self.centralWidget.setLayout(self.vlayout)

        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())
                

class ObjectDetectionResultsWindow(QMainWindow, Window):
    def __init__(self, parent,parentwindow):
        super(ObjectDetectionResultsWindow, self).__init__(parent)
        self.parent = parent
        self.parentwindow = parentwindow
        self.setupUi(self)
        
            
    def saveObjects(self, writer, images, labels):

        if self.parent.TrackingModeEnabled:
            self.saveTrackResults(writer, images, labels)
        else:
            self.saveDetectionResults(writer, images, labels)

    def saveDetectionResults(self, writer, images, labels):
        header = ['image name'] + ['frame number'] + ['number of objects'] + ['object type']
        writer.writerow(header)
        num = len(labels)
        self.parent.emitinitProgress(num)
        
        for image, label in zip(images,labels):
            name = self.parent.files.getFilenameFromPath(self.parent.files.convertIfStackPath(image),withfileextension=True)
            points,_ = loadPoints(self.parent.files.convertIfStackPath(label))
            if self.parent.files.isStackLabel(label):
                frame = self.parent.files.getFrameNumber(label) + 1
            else:
                frame = 1
            pointlabels = [x.classlabel for x in points]
            
            for i in range(1,max(pointlabels)+1):
                row = [name] + [frame] + [pointlabels.count(i)] + [i]
                writer.writerow(row) 

    def saveTrackResults(self, writer, images, labels):
        header = ['object number'] + ['time point'] + ['image name'] + ['frame number'] + ['object type'] + ['position']
        writer.writerow(header)
        num = len(labels)
        self.parent.emitinitProgress(num)

        f1 = lambda x: x.classlabel
        f2 = lambda x: x.getPosition()

        pnt_data = (f1,f2)

        points = [{} for x in range(self.parent.tracking.timepoints)]
        for i,l in zip(images,labels):
            name = self.parent.files.getFilenameFromPath(self.parent.files.convertIfStackPath(i),withfileextension=True)
            point_s,_ = loadPoints(self.parent.files.convertIfStackPath(l))

            if self.parent.tracking.stackMode and self.parent.files.isStackLabel(l):
                tp = self.parent.files.getFrameNumber(l)
            else:
                tp = self.parent.tracking.getTimePointFromImageName(self.parent.files.convertIfStackPath(i))
            if self.parent.files.isStackLabel(l):
                frame = self.parent.files.getFrameNumber(l) + 1
            else:
                frame = 1
            points[tp] = {x.objectNumber: [name] + [frame] + [f(x) for f in pnt_data] for x in point_s}
            self.parent.emitProgress()

        for n in self.parent.tracking.getNumbersOfTrackedObjects():
            for t in self.parent.tracking.getObjectOccurences(n):
                try:
                    row = [n] + [t+1] + points[t][n]
                    writer.writerow(row) 
                except:
                    pass


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 27 16:38:48 2019

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

from abc import ABC, abstractmethod
from enum import Enum
import cv2
import numpy as np
import threading

import time
import utils.shapes.Contour as Contour

class canvasTool(Enum):
    drag = 'drag'
    draw = 'draw'
    assign = 'assign'
    delete = 'delete'
    extend = 'extend'
    poly =  'poly'
    assist = 'assist'
    scale =  'scale'
    shift = 'shift'
    point = 'point'
    setimageclass = 'setimageclass'
    assignimageclass = 'assignimageclass'
    objectnumber = 'objectnumber'
    objectcolor = 'objectcolor'
    deleteobject = 'deleteobject'

class canvasToolButton(Enum):
    # Buttons are set visible or invisible depending on mode
    drag = canvasTool.drag
    draw = canvasTool.draw
    poly =  canvasTool.poly
    point = canvasTool.point
    shift = canvasTool.shift
    assign = canvasTool.assign
    extend = canvasTool.extend
    delete = canvasTool.delete
    assignimageclass = canvasTool.assignimageclass
    setimageclass = canvasTool.setimageclass
    
    

def validTool(f):
    # should decorate all functions that use canvas.painter 
    # to check if a valid tool is taken for that painter
    def wrapped_f(self, *args, **kwargs):
        if self.type in self.canvas.painter.tools:
            return f(self, *args, **kwargs)
        else:
            raise TypeError ('not a valid tool for that painter')
    return wrapped_f


class AbstractTool(ABC):
    def __init__(self, canvas):
        self.canvas = canvas
        self.Text = None
        self.type = None
        super().__init__()
        
    @abstractmethod
    def mouseMoveEvent(self, e):
        pass
    @abstractmethod
    def mouseReleaseEvent(self,e):
        pass
    @abstractmethod
    def mousePressEvent(self,e):
        pass

    def wheelEvent(self,e):
        pass

    @abstractmethod
    def Cursor(self):
        pass
    
    def initialize(self):
        pass

    def ShowSettings(self):
        pass

    def HideSettings(self):
        pass


class DummyTool(AbstractTool):
    # dummy tool, that executes on Button press,
    # no interaction with canvas
    def __init__(self, canvas, text):
        super().__init__(canvas)
        self.Text = text
        self.canvas.setnewTool(canvasTool.drag.name)
        
    def mouseMoveEvent(self, e):
        pass
    def mouseReleaseEvent(self,e):
        pass
    def mousePressEvent(self,e):
        pass
    def Cursor(self):
        return Qt.ArrowCursor  
    
class DragTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Drag"
        self.type = canvasTool.drag
    
    def mouseMoveEvent(self, e):
        pass

    def mouseReleaseEvent(self, e):
        if self.canvas.dragMode() == QGraphicsView.ScrollHandDrag:
            self.canvas.setDragMode(QGraphicsView.NoDrag)
            self.canvas.setCursor(Qt.OpenHandCursor)
            
    def mousePressEvent(self, e):
        if self.canvas.hasImage():
            self.canvas.setDragMode(QGraphicsView.ScrollHandDrag)
                        

    def wheelEvent(self,e):
        if e.angleDelta().y() > 0:
            self.canvas.zoomIn()
        else:
            self.canvas.zoomOut()
                
    def Cursor(self):
        return Qt.OpenHandCursor
                
class DrawTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Draw"
        self.type = canvasTool.draw
        self.drawimage = None

    @property
    def mode(self):
        # 1 add, -1 del, 0 slice
        if self.canvas.parent.CBAddShape.isChecked():
            return 1
        elif self.canvas.parent.CBDelShape.isChecked():
            return -1
        elif self.canvas.parent.CBSliceShape.isChecked():
            return 0

    def __del__(self): 
        if self.canvas.painter.NewContour and self.canvas.painter.NewContour.numPoints() > 1 and self.drawimage is not None:
            self.canvas.copyRect(self.drawimage, self.canvas.image(),self.canvas.getActiveDrawingRect())
            self.drawimage = None            
   

    @validTool
    def mouseMoveEvent(self, e):
        if self.canvas.painter.NewContour:
            p = self.canvas.f2intPoint(self.canvas.mapToScene(e.pos()))
            if not self.canvas.fastPainting and self.mode != 0:
                if self.drawimage is None:
                    self.drawimage = self.canvas.image().copy()
                if self.canvas.painter.NewContour.numPoints() > 1:
                    self.canvas.copyRect(self.drawimage, self.canvas.image(),self.canvas.getActiveDrawingRect(p))
       
                p2 = self.canvas.np2QPoint(self.canvas.painter.NewContour.getLastPoint())
                self.canvas.painter.addPoint2NewContour(p)
                
                self.canvas.copyRect(self.canvas.image(),self.drawimage, self.canvas.getActiveDrawingRect())

                p3 = self.canvas.np2QPoint(self.canvas.painter.NewContour.getFirstPoint())              
                self.canvas.painter.addline(p,p3, dashed = True)
            else:
                self.canvas.painter.addPoint2NewContour(p)
           
    @validTool
    def mouseReleaseEvent(self, e):       
        if self.canvas.painter.NewContour:
            self.canvas.painter.finishNewContour(delete = self.mode < 1, close = self.mode != 0, drawaspolygon = self.mode == 0)
            self.drawimage = None
    
    @validTool            
    def mousePressEvent(self, e):
        if e.button() == Qt.LeftButton:
            if self.canvas.painter.NewContour is None:
                self.canvas.painter.prepareNewContour()
                self.canvas.painter.NewContour = Contour.Contour(self.canvas.parent.activeClass(), self.canvas.QPoint2np(self.canvas.f2intPoint(self.canvas.mapToScene(e.pos()))))

        if e.button() == Qt.RightButton:
            if self.mode == 1:
                self.canvas.parent.CBDelShape.setChecked(True) 
            elif self.mode == 0:
                self.canvas.parent.CBAddShape.setChecked(True)
            elif self.mode == -1:
                self.canvas.parent.CBSliceShape.setChecked(True)
            self.canvas.setCursor(self.Cursor())
       
    def Cursor(self):
        return Qt.ArrowCursor

    def ShowSettings(self):
        self.canvas.parent.DrawSettings.show()

    def HideSettings(self):
        self.canvas.parent.DrawSettings.hide()
    
    
class PolygonTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Polygon"
        self.type = canvasTool.poly
        self.drawimage = None
        self.lock = threading.Lock()
        
    def __del__(self): 
        if self.canvas.painter.NewContour and self.canvas.painter.NewContour.numPoints() > 1 and self.drawimage is not None:
            self.canvas.copyRect(self.drawimage, self.canvas.image(),self.canvas.getFieldOfViewRect())
            self.drawimage = None    
        
    @property
    def mode(self):
        # 1 add, -1 del, 0 slice
        if self.canvas.parent.CBAddShape.isChecked():
            return 1
        elif self.canvas.parent.CBDelShape.isChecked():
            return -1
        elif self.canvas.parent.CBSliceShape.isChecked():
            return 0

    @validTool
    def mouseMoveEvent(self, e):
        if self.canvas.painter.NewContour is not None:
            if not self.canvas.fastPainting:
                p2 = self.canvas.f2intPoint(self.canvas.mapToScene(e.pos()))
                if self.drawimage is None:
                    self.drawimage = self.canvas.image().copy()
                self.canvas.copyRect(self.drawimage, self.canvas.image(),self.canvas.getFieldOfViewRect())
                p1 = self.canvas.np2QPoint(self.canvas.painter.NewContour.getLastPoint())
                self.canvas.painter.addline(p1,p2, dashed = True)              
    
    @validTool
    def mouseReleaseEvent(self,e):
        if e.button() == Qt.RightButton:
            if self.canvas.painter.NewContour:
                self.canvas.painter.finishNewContour(delete = self.mode < 1, close = self.mode != 0, drawaspolygon = self.mode == 0)
                self.drawimage = None
            else:
                if self.mode == 1:
                    self.canvas.parent.CBDelShape.setChecked(True) 
                elif self.mode == 0:
                    self.canvas.parent.CBAddShape.setChecked(True)
                elif self.mode == -1:
                    self.canvas.parent.CBSliceShape.setChecked(True)
                self.canvas.setCursor(self.Cursor())

        elif e.button() == Qt.LeftButton:
            if self.canvas.painter.NewContour is None:
                self.canvas.painter.prepareNewContour()
                self.canvas.painter.NewContour = Contour.Contour(self.canvas.parent.activeClass(), self.canvas.QPoint2np(self.canvas.f2intPoint(self.canvas.mapToScene(e.pos()))))
            else:
                p = self.canvas.f2intPoint(self.canvas.mapToScene(e.pos()))
                self.canvas.painter.addPoint2NewContour(p)
                if self.drawimage is None:
                    self.drawimage = self.canvas.image().copy()
                self.canvas.copyRect(self.canvas.image(),self.drawimage, self.canvas.getFieldOfViewRect())
     
    @validTool                 
    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.ArrowCursor

    def ShowSettings(self):
        self.canvas.parent.DrawSettings.show()

    def HideSettings(self):
        self.canvas.parent.DrawSettings.hide()
    
class AssignTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Assign"
        self.type = canvasTool.assign

    def mouseMoveEvent(self, e):
        pass
        
    @validTool
    def mouseReleaseEvent(self,e):
        if e.button() == Qt.LeftButton:
            contour = self.canvas.painter.shapes.getShape(self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())),self.canvas.FontSize)
            if contour is not None:
                contour.setClassLabel(self.canvas.parent.activeClass())
                self.canvas.painter.checkForChanges()
                
    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.ArrowCursor

        
class DeleteTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Delete"
        self.type = canvasTool.delete
        
    def mouseMoveEvent(self, e):
        pass
        
    @validTool
    def mouseReleaseEvent(self,e):
        if e.button() == Qt.LeftButton:
            
            contour = self.canvas.painter.shapes.getShapeOfClass_x(self.canvas.parent.activeClass(),self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())), self.canvas.FontSize)
            if contour is None:
                contour = self.canvas.painter.shapes.getShape(self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())), self.canvas.FontSize)
            if contour is not None:
                self.canvas.painter.shapes.deleteShape(contour)
                self.canvas.redrawImage()
                      
    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.ArrowCursor
        
    
class ExtendTool(AbstractTool):
    
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Extend/Erase"
        self.type = canvasTool.extend
        self.inprogress = False
        self.lock = threading.RLock()
        
    @validTool
    def mouseMoveEvent(self, e):
        if self.canvas.painter.sketch is None:
            return
        with self.lock:
            # if e.button() == Qt.LeftButton: is not working here for some reasons
            if self.inprogress:
                p0 = self.canvas.mapToScene(e.pos())
                p = self.canvas.QPoint2np(p0)
                x = p[0,0]
                y = p[0,1]
                # self.canvas.painter.addCircle(p0, self.canvas.parent.SSize.value(), erase = self.canvas.parent.CBErase.isChecked())
                self.canvas.painter.addCircle(p0, self.canvas.parent.SSize.value())
                val = 1 if self.canvas.parent.CBErase.isChecked() == False else 0
                cv2.circle(self.canvas.painter.sketch, (x, y), self.canvas.parent.SSize.value(), (val), -1)

    @validTool 
    def mouseReleaseEvent(self,e):
        if e.button() == Qt.LeftButton:
            self.canvas.painter.finishNewContour()
            self.inprogress = False

    @validTool                 
    def mousePressEvent(self,e):
        if self.canvas.image() is None:
            return
        if e.button() == Qt.LeftButton:  
            self.size = self.canvas.parent.SSize.value()
            self.canvas.painter.prepareNewContour()
            self.inprogress = True
       
        elif e.button() == Qt.RightButton:
            self.canvas.parent.CBExtend.setChecked(True) if self.canvas.parent.CBErase.isChecked() else self.canvas.parent.CBErase.setChecked(True)
            self.canvas.setCursor(self.Cursor())
    
    @validTool
    def wheelEvent(self,e):
        if e.angleDelta().y() > 0:
            v = self.canvas.parent.SSize.value() + 1
            self.canvas.parent.SSize.setValue(v)
        else:
            v = self.canvas.parent.SSize.value() - 1
            self.canvas.parent.SSize.setValue(v)
        self.size = self.canvas.parent.SSize.value()
        self.canvas.setCursor(self.Cursor())

    def createCursor(self):
        if not self.canvas.hasImage() or not self.canvas.zoomfactor():
            return Qt.ArrowCursor
        
        size = (2*self.canvas.parent.SSize.value() + self.canvas.pen_size) * self.canvas.zoomfactor()[0]
        # unfortunately max size is somewhere around 100
        if size > 95:
            return Qt.SizeAllCursor
        icon = QPixmap(size+1,size+1)
        icon.fill(Qt.transparent)
        p = QPainter(icon)
        color = self.canvas.parent.ClassColor()
        if self.canvas.parent.CBErase.isChecked():
            p.setBrush(QBrush(Qt.NoBrush))
            p.setPen(QPen(color,Qt.SolidLine))
        else:
            p.setBrush(QBrush(color,Qt.SolidPattern))
            p.setPen(QPen(Qt.NoPen))
            
        p.drawEllipse(0, 0, size, size)
        p.end()
        return QCursor(icon, -1, -1)

    def Cursor(self):
        return self.createCursor()
    
    def ShowSettings(self):
        self.canvas.parent.ExtendSettings.show()

    def HideSettings(self):
        self.canvas.parent.ExtendSettings.hide()


class DEXTRTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = 'Assisted Segmentation'
        self.type = canvasTool.assist
        self.extremePoints = []

    def __del__(self): 
        self.extremePoints.clear()
        self.canvas.redrawImage()
        
    def mouseMoveEvent(self, e):
        pass

    @validTool  
    def mouseReleaseEvent(self,e):
        if self.canvas.image() is None:
            return
        if e.button() == Qt.LeftButton:  
            p = self.canvas.mapToScene(e.pos())

            h = self.canvas.pen_size
            w = self.canvas.pen_size
            x = p.x() - w//2
            y = p.y() - h//2

            self.canvas.painter.addRectangle(x, y, h, w)
            self.extremePoints.append(self.canvas.QPoint2npPoint(self.canvas.f2intPoint(p)))
            if len(self.extremePoints) == 4:
                self.canvas.painter.assistedSegmentation(self.extremePoints)
                self.extremePoints.clear()

    def mousePressEvent(self,e):
        pass
    def Cursor(self):
        return Qt.ArrowCursor  

        
class PointTool(AbstractTool):
    
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Point"
        self.type = canvasTool.point
        
    def mouseMoveEvent(self, e):
        pass

     
    @validTool   
    def mouseReleaseEvent(self,e):
        if self.canvas.image() is None:
            return
        if e.button() == Qt.LeftButton:
            p = self.canvas.f2intPoint(self.canvas.mapToScene(e.pos()))
            self.canvas.painter.addPoint(p)
            
        elif e.button() == Qt.RightButton:
            self.canvas.setnewTool(canvasTool.delete.name)

          
    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.CrossCursor      
    
class ShiftTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Shift"
        self.type = canvasTool.shift
        self.shiftedShape = None
        self.shifting = False
        self.origin_coordinates = None

    @validTool    
    def mouseMoveEvent(self, e):
        if self.shiftedShape and self.shifting:
            p = self.canvas.mapToScene(e.pos())
            self.shiftedShape.coordinates = self.canvas.QPoint2npPoint(self.canvas.f2intPoint(p))
            if not self.canvas.fastPainting:
                self.canvas.redrawImage(True)

    @validTool    
    def mouseReleaseEvent(self,e):
        if not self.shiftedShape:
            return
        self.shifting = False

        self.shiftedShape = None
        self.canvas.redrawImage()

    @validTool     
    def mousePressEvent(self,e):
        if self.canvas.image() is None:
            return
        p = self.canvas.mapToScene(e.pos())

        shape = self.canvas.painter.shapes.getShapeOfClass_x(self.canvas.parent.activeClass(),self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())), self.canvas.FontSize)
        if shape is None:
            shape = self.canvas.painter.shapes.getShape(self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())), self.canvas.FontSize)
        if shape is not None:
            self.shifting = True
            self.shiftedShape = shape
            self.origin_coordinates = shape.coordinates


    def Cursor(self):
        return Qt.SizeAllCursor   


class ScaleTool(AbstractTool):
    
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Scale"
        self.type = canvasTool.scale
        self.size = 15
        self.color = QColor(0,0,255)
        self.num = 0
        self.p1 = None
        self.dist_inpixel = 0
        
    def mouseMoveEvent(self, e):
        pass

        
    def mouseReleaseEvent(self,e):
        if self.canvas.image() is None:
            return
        if e.button() == Qt.LeftButton:
            if self.num == 0:
                self.p1 = self.canvas.mapToScene(e.pos())
                self.canvas.painter.addTriangle(self.p1, self.size, self.color)
                self.num += 1
            else:  
                p2 = self.canvas.mapToScene(e.pos())
                self.canvas.painter.addTriangle(p2, self.size, self.color)
                dist_inpixel = abs(self.p1.x() - p2.x())
                dist_inum,ok = QInputDialog.getInt(self.canvas.parent, "Set Scale","enter distance in \u03BCm")
                if ok and dist_inum > 0:
                    self.canvas.setScale(dist_inpixel / dist_inum * 1000)
                
                self.canvas.setnewTool(canvasTool.drag.name)
                self.canvas.redrawImage()
                self.num = 0
          
    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.UpArrowCursor
    

# tracking tools
class ObjectNumberTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "ObjectNumber"
        self.type = canvasTool.objectnumber

    def mouseMoveEvent(self, e):
        pass
        
    @validTool
    def mouseReleaseEvent(self,e):
        if e.button() == Qt.LeftButton:
            shape = self.canvas.painter.shapes.getShape(self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())),self.canvas.FontSize)
            if shape is not None:
                objnum,ok = QInputDialog.getInt(self.canvas.parent, "Set Object Number","Object Number", shape.objectNumber, 1, 100000)
                if ok and objnum > 0:
                    old_obj = self.canvas.painter.shapes.getShapeNumberX(objnum)
                    if old_obj:
                        old_obj.objectNumber = self.canvas.parent.tracking.getFirstFreeObjectNumber()
                    oldobjnum = self.canvas.parent.tracking.setObjectNumber(obj=shape, objnum=objnum)
                    self.canvas.redrawImage()
                     
    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.ArrowCursor

class ObjectColorTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "ObjectColor"
        self.type = canvasTool.objectcolor

    def mouseMoveEvent(self, e):
        pass
        
    @validTool
    def mouseReleaseEvent(self,e):
        if e.button() == Qt.LeftButton:
            shape = self.canvas.painter.shapes.getShape(self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())),self.canvas.FontSize)
            if shape is not None:
                color = QColorDialog.getColor()
                if color.isValid():
                    self.canvas.painter.objectcolors[shape.objectNumber-1] = color
                    self.canvas.redrawImage()

    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.ArrowCursor

class DeleteObjectTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "DeleteObject"
        self.type = canvasTool.deleteobject

    def mouseMoveEvent(self, e):
        pass
        
    @validTool
    def mouseReleaseEvent(self,e):
        if e.button() == Qt.LeftButton:
            shape = self.canvas.painter.shapes.getShape(self.canvas.QPoint2np(self.canvas.mapToScene(e.pos())),self.canvas.FontSize)
            if shape is not None:
                with self.canvas.parent.wait_cursor():
                    self.canvas.parent.tracking.deleteObject(shape)
                    self.canvas.painter.shapes.deleteShape(shape)
                    self.canvas.redrawImage()
  
    def mousePressEvent(self,e):
        pass

    def Cursor(self):
        return Qt.ArrowCursor
    

# classification tools
class setImageClassTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Set image class"
        self.type = canvasTool.setimageclass
        
    def initialize(self):
        # a bit hacky and poor designed as this is the execute
        self.canvas.assignImageClass()
        self.canvas.parent.nextImage()
        self.canvas.setnewTool(canvasTool.drag.name)

    def mouseReleaseEvent(self,e):
        pass
    def mouseMoveEvent(self,e):
        pass
    def mousePressEvent(self,e):
        pass
    def Cursor(self):
        return Qt.ArrowCursor  
        
        
class assignImageClassTool(AbstractTool):
    def __init__(self, canvas):
        super().__init__(canvas)
        self.canvas = canvas
        self.Text = "Assign image class"
        self.type = canvasTool.assignimageclass
        
    def initialize(self):
        self.canvas.assignImageClass()
        self.canvas.setnewTool(canvasTool.drag.name)
        
    def mouseReleaseEvent(self,e):
        pass
    def mouseMoveEvent(self,e):
        pass
    def mousePressEvent(self,e):
        pass
    def Cursor(self):
        return Qt.ArrowCursor  


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 14 16:56:31 2020

@author: Koerber
"""

from scipy.optimize import linear_sum_assignment
import numpy as np
import math
from utils.Image import readNumOfImageFrames



class ObjectTracking():
    def __init__(self,dl,files):
        self.dl = dl
        self.files = files
        self.tracking_list = []
        self.thresh = 35
        self.fadeawayperiod = 1
        self.newobjectperiod = 2
        self.objects = None
        self.timepoints = 0
        self.sequence = []
        self.stackMode = False
        # results
        self.tracks = None
          
    def deleteObject(self, obj):
        objnum = obj.objectNumber
        for tp,t in self.labelGenerator():
            shapes = self.dl.Mode.LoadShapes(t)
            numbers = [x.objectNumber for x in shapes]
            if objnum in numbers:
                shapes.remove(shapes[numbers.index(objnum)])
                self.dl.Mode.saveShapes(shapes, t)
        np.delete(self.tracks, objnum-1,0)

    def changeTimePoint(self, objects, tp):
        self.tracks[:,tp,:] = -1
        unmatched = []
        for o in objects.shapes:
            if o.objectNumber == -1 or self.tracks.shape[0] < o.objectNumber:
                unmatched.append(o)
            else:
                self.tracks[o.objectNumber-1,tp,:] = np.asarray(o.getPosition())
        if len(unmatched) == 0:
            return

        if tp != 0:
            t_minus_one = self.tracks[:,tp-1,:]
            t = self.tracks[:,tp,:]
            lost = np.where((t_minus_one > 0) & (t < 0))
            if lost[0].size == 0:
                for o in unmatched:
                    self.setObjectNumber(o,tp)
            else:
                pos = [x for x in zip(t_minus_one[lost][0::2],t_minus_one[lost][1::2])]
                corr_mat = [self.distance(x.getPosition(),y) for x in unmatched for y in pos]
                mat = np.asarray(corr_mat).reshape(len(unmatched),len(lost[0])//2)
                row,col = linear_sum_assignment(mat)
                for r,c in zip(row,col):
                    if self.stackMode:
                        frames = self.files.ImagePath2LabelPath(self.sequence[0], False, True)
                        prev_shapes = self.dl.Mode.LoadShapes(frames[[self.files.getFrameNumber(x) for x in frames].index(tp-1)])
                    else:
                        prev_shapes = self.dl.Mode.LoadShapes(self.files.ImagePath2LabelPath(self.sequence[tp-1]))
                    prev_shape = prev_shapes[[x.objectNumber for x in prev_shapes].index(lost[0][c*2]+1)]
                    if mat[r,c] < self.thresh and unmatched[r].classlabel == prev_shape.classlabel:
                        self.setObjectNumber(unmatched[r],tp,lost[0][c*2] + 1)
                    else:
                        self.setObjectNumber(unmatched[r],tp)

                for o in range(len(unmatched)):
                    if o not in row:
                        self.setObjectNumber(unmatched[o],tp)
        else:
            for o in unmatched:
                self.setObjectNumber(o,tp)

    def labelGenerator(self):
        if not self.files.files:
            return
        for tp in range(len(self.files.files)):
            imagefile = self.sequence[tp]
            t = self.files.ImagePath2LabelPath(imagefile, True, self.stackMode)
            if not t:
                continue
            if self.stackMode:
                for t_ in t:
                    yield tp + self.files.getFrameNumber(t_), t_
            else:
                yield tp, t

    def setObjectNumber(self, obj, tp=None, objnum=None):
        if objnum is None:
            objnum = self.getFirstFreeObjectNumber()
        if tp is None:
            tp = self.files.currentImage
        obj.objectNumber = objnum
        self.addTrackPosition(tp, obj)

    def getFirstFreeObjectNumber(self):
        free = np.where(np.all(self.tracks<0,axis=1))[0]
        if len(free) > 0:
            return free[0] + 1
        else:
            return self.tracks.shape[0]

    def distance(self, p1, p2):
        return math.sqrt((p1[0]-p2[0])**2 + (p1[1]-p2[1])**2)

    def getObjectTrack(self, objnum):
        if self.tracks.shape[0]>=objnum and objnum > 0:   
            return self.tracks[objnum-1,...].tolist()
        else:
            return None

    def getObjectOccurences(self, objnum):
        if self.tracks.shape[0]>=objnum and objnum > 0:   
            return np.where(self.tracks[objnum-1,...,0]>0)[0].tolist()
        else:
            return []

    def getTimePointFromImageName(self, imagepath):
        return self.sequence.index(imagepath)

    def getNumbersOfTrackedObjects(self):
        return np.where(np.any(self.tracks[...,0]>=0,axis=1))[0]+1

    def resetTracking(self, numofTimePoints, numofObjects=250):
        self.tracks = np.zeros((numofObjects,numofTimePoints,2),dtype=int)-1
        self.objects = None
        self.tracking_list = []

    def loadTrack(self):
        # load from files
        if not self.files.files:
            return
        self.sequence = self.files.files
        if len(self.files.files) > 1:
            self.timepoints = len(self.files.files)
            self.stackMode = False
        elif len(self.files.files) == 1:
            self.timepoints = readNumOfImageFrames(self.files.files[0])
            self.stackMode = True

        self.resetTracking(self.timepoints)
        for tp,t in self.labelGenerator():
            shapes = self.dl.Mode.LoadShapes(t)
            save = False
            for i,s in enumerate(shapes):
                # fix object number
                # that would preferably done with shape object 
                if s.objectNumber == -1:
                    s.objectNumber = i+1
                    save = True
                self.addTrackPosition(tp,s)
            if save:
                self.dl.Mode.saveShapes(shapes, t)
            self.fillBlanksForMissingObjects(tp)

    
    #def saveTrack(self, filename):
    # could be saved for performance reasons instead of reloaded from individual contour files
    #    pass

    def addTrackPosition(self, tp, obj):
        objnum = obj.objectNumber
        if self.tracks.shape[0] < objnum+1:
            self.tracks = np.append(self.tracks,np.zeros((objnum - self.tracks.shape[0],self.tracks.shape[1],2))-1,axis = 0)
        pos = obj.getPosition()
        self.tracks[objnum-1,tp,:] = np.asarray(pos)


    def fillBlanksForMissingObjects(self, tp):      
        [x.append((-1,-1)) for x in self.tracks if len(x) < tp+1]


    def performTracking(self):
        self.resetTracking(self.timepoints)
        for tp,t in self.labelGenerator():
            shapes = self.dl.Mode.LoadShapes(t)
            self.addTimePoint(self.dl.Mode.LoadShapes(t), tp)
            for i in range(self.objects.shape[0]):
                if self.objects[i,tp] > 0:
                    shapes[self.objects[i,tp]-1].setObjectNumber(i+1) 
                    self.addTrackPosition(tp,shapes[self.objects[i,tp]-1])
            self.fillBlanksForMissingObjects(tp)
            self.dl.Mode.saveShapes(shapes, t)

    def addTimePoint(self, detections, t):
        #if t == 0:
        if self.objects is None or self.tracking_list == []:
            self.objects = np.zeros((len(detections),self.timepoints),dtype=int)
            self.objects[:,0] = range(1,len(detections)+1)
            self.tracking_list = [(x,i) for (x,i) in zip(detections,range(1,len(detections)+1))]
        else:
            self.calculateMatches(detections, t)
            

    def calculateMatches(self,  t, tp):
        t_minus_one = [x for (x,i) in self.tracking_list]
        corr_mat = [self.dl.Mode.LabelDistance(x,y) for x in t_minus_one for y in t]
        mat = np.asarray(corr_mat).reshape(len(t_minus_one),len(t))          
         
        # improves performance because it avoid (or equalizes to be more precisely) matches larger self.thresh
        mat[mat>self.thresh] = 100000
        row,col = linear_sum_assignment(mat)

        # matched     
        for r,c in zip(row,col):
            if mat[r,c] < self.thresh and t_minus_one[r].classlabel == t[c].classlabel: 
                pos = self.tracking_list[r][1]-1
                # update contour
                self.tracking_list[r] = (t[c], self.tracking_list[r][1])
                self.objects[pos,tp] = c+1
            else:
                newobject = np.zeros((1,self.timepoints),dtype=int)
                newobject[0,tp] = c+1
                self.objects = np.append(self.objects,newobject, axis=0)
                self.tracking_list.append((t[c], self.objects.shape[0]))
                
        # unmatched
        for track in range(len(t)):
            if track not in col:
                newobject = np.zeros((1,self.timepoints),dtype=int)
                newobject[0,tp] = track+1
                self.objects = np.append(self.objects,newobject, axis=0)
                self.tracking_list.append((t[track], self.objects.shape[0]))

           
        # remove continuously undetected 
        gone = []
        if tp-self.fadeawayperiod >= 0:
            for x in self.tracking_list:
                if self.objects[x[1]-1,tp] == 0:
                    occ = self.objects[x[1]-1,tp-self.fadeawayperiod:tp]
                    if np.sum(occ) == 0:
                        gone.append(x)    
            for obj in gone:
                self.tracking_list.remove(obj)

        




# Copyright 2019 The TensorFlow Authors, Pavel Yakubovskiy, Björn Barz. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains definitions for EfficientNet model.

[1] Mingxing Tan, Quoc V. Le
  EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks.
  ICML'19, https://arxiv.org/abs/1905.11946
"""

# Code of this model implementation is mostly written by
# Björn Barz ([@Callidior](https://github.com/Callidior))

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import json
import math
import string
import collections
import numpy as np

from six.moves import xrange
from dl.models.keras_applications.imagenet_utils import _obtain_input_shape
from dl.models.keras_applications.imagenet_utils import decode_predictions
from dl.models.keras_applications.imagenet_utils import preprocess_input as _preprocess_input

from . import get_submodules_from_kwargs

from dl.utils.dl_downloads import get_EfficientNet

backend = None
layers = None
models = None
keras_utils = None


BlockArgs = collections.namedtuple('BlockArgs', [
    'kernel_size', 'num_repeat', 'input_filters', 'output_filters',
    'expand_ratio', 'id_skip', 'strides', 'se_ratio'
])
# defaults will be a public argument for namedtuple in Python 3.7
# https://docs.python.org/3/library/collections.html#collections.namedtuple
BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)

DEFAULT_BLOCKS_ARGS = [
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=32, output_filters=16,
              expand_ratio=1, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=2, input_filters=16, output_filters=24,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=2, input_filters=24, output_filters=40,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=3, input_filters=40, output_filters=80,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=3, input_filters=80, output_filters=112,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25),
    BlockArgs(kernel_size=5, num_repeat=4, input_filters=112, output_filters=192,
              expand_ratio=6, id_skip=True, strides=[2, 2], se_ratio=0.25),
    BlockArgs(kernel_size=3, num_repeat=1, input_filters=192, output_filters=320,
              expand_ratio=6, id_skip=True, strides=[1, 1], se_ratio=0.25)
]

CONV_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 2.0,
        'mode': 'fan_out',
        # EfficientNet actually uses an untruncated normal distribution for
        # initializing conv layers, but keras.initializers.VarianceScaling use
        # a truncated distribution.
        # We decided against a custom initializer for better serializability.
        'distribution': 'normal'
    }
}

DENSE_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 1. / 3.,
        'mode': 'fan_out',
        'distribution': 'uniform'
    }
}


def preprocess_input(x, **kwargs):
    kwargs = {k: v for k, v in kwargs.items() if k in ['backend', 'layers', 'models', 'utils']}
    return _preprocess_input(x, mode='torch', **kwargs)


def get_swish(**kwargs):
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)
    def swish(x):
        """Swish activation function: x * sigmoid(x).
        Reference: [Searching for Activation Functions](https://arxiv.org/abs/1710.05941)
        """

        if backend.backend() == 'tensorflow':
            try:
                # The native TF implementation has a more
                # memory-efficient gradient implementation
                return backend.tf.nn.swish(x)
            except AttributeError:
                pass

        return x * backend.sigmoid(x)
    return  swish


def get_dropout(**kwargs):
    """Wrapper over custom dropout. Fix problem of ``None`` shape for tf.keras.
    It is not possible to define FixedDropout class as global object,
    because we do not have modules for inheritance at first time.

    Issue:
        https://github.com/tensorflow/tensorflow/issues/30946
    """
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    class FixedDropout(layers.Dropout):
        def _get_noise_shape(self, inputs):
            if self.noise_shape is None:
                return self.noise_shape

            symbolic_shape = backend.shape(inputs)
            noise_shape = [symbolic_shape[axis] if shape is None else shape
                           for axis, shape in enumerate(self.noise_shape)]
            return tuple(noise_shape)

    return FixedDropout


def round_filters(filters, width_coefficient, depth_divisor):
    """Round number of filters based on width multiplier."""

    filters *= width_coefficient
    new_filters = int(filters + depth_divisor / 2) // depth_divisor * depth_divisor
    new_filters = max(depth_divisor, new_filters)
    # Make sure that round down does not go down by more than 10%.
    if new_filters < 0.9 * filters:
        new_filters += depth_divisor
    return int(new_filters)


def round_repeats(repeats, depth_coefficient):
    """Round number of repeats based on depth multiplier."""

    return int(math.ceil(depth_coefficient * repeats))


def mb_conv_block(inputs, block_args, activation, drop_rate=None, prefix='', ):
    """Mobile Inverted Residual Bottleneck."""

    has_se = (block_args.se_ratio is not None) and (0 < block_args.se_ratio <= 1)
    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    # workaround over non working dropout with None in noise_shape in tf.keras
    Dropout = get_dropout(
        backend=backend,
        layers=layers,
        models=models,
        utils=keras_utils
    )

    # Expansion phase
    filters = block_args.input_filters * block_args.expand_ratio
    if block_args.expand_ratio != 1:
        x = layers.Conv2D(filters, 1,
                          padding='same',
                          use_bias=False,
                          kernel_initializer=CONV_KERNEL_INITIALIZER,
                          name=prefix + 'expand_conv')(inputs)
        x = layers.BatchNormalization(axis=bn_axis, name=prefix + 'expand_bn')(x)
        x = layers.Activation(activation, name=prefix + 'expand_activation')(x)
    else:
        x = inputs

    # Depthwise Convolution
    x = layers.DepthwiseConv2D(block_args.kernel_size,
                               strides=block_args.strides,
                               padding='same',
                               use_bias=False,
                               depthwise_initializer=CONV_KERNEL_INITIALIZER,
                               name=prefix + 'dwconv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name=prefix + 'bn')(x)
    x = layers.Activation(activation, name=prefix + 'activation')(x)

    # Squeeze and Excitation phase
    if has_se:
        num_reduced_filters = max(1, int(
            block_args.input_filters * block_args.se_ratio
        ))
        se_tensor = layers.GlobalAveragePooling2D(name=prefix + 'se_squeeze')(x)

        target_shape = (1, 1, filters) if backend.image_data_format() == 'channels_last' else (filters, 1, 1)
        se_tensor = layers.Reshape(target_shape, name=prefix + 'se_reshape')(se_tensor)
        se_tensor = layers.Conv2D(num_reduced_filters, 1,
                                  activation=activation,
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_reduce')(se_tensor)
        se_tensor = layers.Conv2D(filters, 1,
                                  activation='sigmoid',
                                  padding='same',
                                  use_bias=True,
                                  kernel_initializer=CONV_KERNEL_INITIALIZER,
                                  name=prefix + 'se_expand')(se_tensor)
        if backend.backend() == 'theano':
            # For the Theano backend, we have to explicitly make
            # the excitation weights broadcastable.
            pattern = ([True, True, True, False] if backend.image_data_format() == 'channels_last'
                       else [True, False, True, True])
            se_tensor = layers.Lambda(
                lambda x: backend.pattern_broadcast(x, pattern),
                name=prefix + 'se_broadcast')(se_tensor)
        x = layers.multiply([x, se_tensor], name=prefix + 'se_excite')

    # Output phase
    x = layers.Conv2D(block_args.output_filters, 1,
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name=prefix + 'project_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name=prefix + 'project_bn')(x)
    if block_args.id_skip and all(
            s == 1 for s in block_args.strides
    ) and block_args.input_filters == block_args.output_filters:
        if drop_rate and (drop_rate > 0):
            x = Dropout(drop_rate,
                        noise_shape=(None, 1, 1, 1),
                        name=prefix + 'drop')(x)
        x = layers.add([x, inputs], name=prefix + 'add')

    return x


def EfficientNet(width_coefficient,
                 depth_coefficient,
                 default_resolution,
                 dropout_rate=0.2,
                 drop_connect_rate=0.2,
                 depth_divisor=8,
                 blocks_args=DEFAULT_BLOCKS_ARGS,
                 model_name='efficientnet',
                 include_top=True,
                 weights='imagenet',
                 input_tensor=None,
                 input_shape=None,
                 pooling=None,
                 classes=1000,
                 **kwargs):
    """Instantiates the EfficientNet architecture using given scaling coefficients.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.
    # Arguments
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        default_resolution: int, default input image size.
        dropout_rate: float, dropout rate before final classifier layer.
        drop_connect_rate: float, dropout rate at skip connections.
        depth_divisor: int.
        blocks_args: A list of BlockArgs to construct block modules.
        model_name: string, model name.
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False.
            It should have exactly 3 inputs channels.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.
    # Returns
        A Keras model instance.
    # Raises
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    if not (weights in {'imagenet', None} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization), `imagenet` '
                         '(pre-training on ImageNet), '
                         'or the path to the weights file to be loaded.')

    if weights == 'imagenet' and include_top and classes != 1000:
        raise ValueError('If using `weights` as `"imagenet"` with `include_top`'
                         ' as true, `classes` should be 1000')

    # Determine proper input shape
    input_shape = _obtain_input_shape(input_shape,
                                      default_size=default_resolution,
                                      min_size=32,
                                      data_format=backend.image_data_format(),
                                      require_flatten=include_top,
                                      weights=weights)

    if input_tensor is None:
        img_input = layers.Input(shape=input_shape)
    else:
        if backend.backend() == 'tensorflow':
            from tensorflow.python.keras.backend import is_keras_tensor
        else:
            is_keras_tensor = backend.is_keras_tensor
        if not is_keras_tensor(input_tensor):
            img_input = layers.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1
    activation = get_swish(**kwargs)

    # Build stem
    x = img_input
    x = layers.Conv2D(round_filters(32, width_coefficient, depth_divisor), 3,
                      strides=(2, 2),
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name='stem_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name='stem_bn')(x)
    x = layers.Activation(activation, name='stem_activation')(x)

    # Build blocks
    num_blocks_total = sum(block_args.num_repeat for block_args in blocks_args)
    block_num = 0
    for idx, block_args in enumerate(blocks_args):
        assert block_args.num_repeat > 0
        # Update block input and output filters based on depth multiplier.
        block_args = block_args._replace(
            input_filters=round_filters(block_args.input_filters,
                                        width_coefficient, depth_divisor),
            output_filters=round_filters(block_args.output_filters,
                                         width_coefficient, depth_divisor),
            num_repeat=round_repeats(block_args.num_repeat, depth_coefficient))

        # The first block needs to take care of stride and filter size increase.
        drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
        x = mb_conv_block(x, block_args,
                          activation=activation,
                          drop_rate=drop_rate,
                          prefix='block{}a_'.format(idx + 1))
        block_num += 1
        if block_args.num_repeat > 1:
            # pylint: disable=protected-access
            block_args = block_args._replace(
                input_filters=block_args.output_filters, strides=[1, 1])
            # pylint: enable=protected-access
            for bidx in xrange(block_args.num_repeat - 1):
                drop_rate = drop_connect_rate * float(block_num) / num_blocks_total
                block_prefix = 'block{}{}_'.format(
                    idx + 1,
                    string.ascii_lowercase[bidx + 1]
                )
                x = mb_conv_block(x, block_args,
                                  activation=activation,
                                  drop_rate=drop_rate,
                                  prefix=block_prefix)
                block_num += 1

    # Build top
    x = layers.Conv2D(round_filters(1280, width_coefficient, depth_divisor), 1,
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name='top_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name='top_bn')(x)
    x = layers.Activation(activation, name='top_activation')(x)
    if include_top:
        x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
        if dropout_rate and dropout_rate > 0:
            x = layers.Dropout(dropout_rate, name='top_dropout')(x)
        x = layers.Dense(classes,
                         activation='softmax',
                         kernel_initializer=DENSE_KERNEL_INITIALIZER,
                         name='probs')(x)
    else:
        if pooling == 'avg':
            x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
        elif pooling == 'max':
            x = layers.GlobalMaxPooling2D(name='max_pool')(x)

    # Ensure that the model takes into account
    # any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = keras_utils.get_source_inputs(input_tensor)
    else:
        inputs = img_input

    # Create model.
    model = models.Model(inputs, x, name=model_name)

    # Load weights.
    if weights == 'imagenet':
        weights_path = get_EfficientNet(model_name, include_top)

        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    return model


def EfficientNetB0(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.0, 1.0, 224, 0.2,
                        model_name='efficientnet-b0',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB1(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.0, 1.1, 240, 0.2,
                        model_name='efficientnet-b1',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB2(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.1, 1.2, 260, 0.3,
                        model_name='efficientnet-b2',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB3(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.2, 1.4, 300, 0.3,
                        model_name='efficientnet-b3',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB4(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.4, 1.8, 380, 0.4,
                        model_name='efficientnet-b4',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB5(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.6, 2.2, 456, 0.4,
                        model_name='efficientnet-b5',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB6(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.8, 2.6, 528, 0.5,
                        model_name='efficientnet-b6',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB7(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(2.0, 3.1, 600, 0.5,
                        model_name='efficientnet-b7',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


setattr(EfficientNetB0, '__doc__', EfficientNet.__doc__)
setattr(EfficientNetB1, '__doc__', EfficientNet.__doc__)
setattr(EfficientNetB2, '__doc__', EfficientNet.__doc__)
setattr(EfficientNetB3, '__doc__', EfficientNet.__doc__)
setattr(EfficientNetB4, '__doc__', EfficientNet.__doc__)
setattr(EfficientNetB5, '__doc__', EfficientNet.__doc__)
setattr(EfficientNetB6, '__doc__', EfficientNet.__doc__)
setattr(EfficientNetB7, '__doc__', EfficientNet.__doc__)



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 13:43:02 2020

@author: Koerber
"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import multiprocessing
import tensorflow as tf

from ui.ui_utils import LabelledSpinBox, LabelledDoubleSpinBox
from ui.style import styleForm

class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Settings') 
        styleForm(Form)

        self.centralWidget = QWidget(Form)
        
        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(3, 10, 3, 3)
        self.vlayout.setSpacing(2)
        self.centralWidget.setLayout(self.vlayout)
        
        self.dlParams = self.DLGroup()
        self.vlayout.addWidget(self.dlParams)
        
        self.dispParams = self.DispGroup()
        self.vlayout.addWidget(self.dispParams)

        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())

        
    def DLGroup(self):
        groupBox = QGroupBox("Deep Learning Settings")
        vlayout = QVBoxLayout(self.centralWidget)
        
        vlayout.setContentsMargins(1, 1, 1, 1)
        vlayout.setSpacing(1)
        
        hlayout = QHBoxLayout(self.centralWidget)
        self.CBgpu = QComboBox(self.centralWidget)
        self.CBgpu.setObjectName('GPUSettings')

        devices = tf.config.experimental.list_physical_devices()
        self.CBgpu.addItem("default")
        for dev in devices:
            self.CBgpu.addItem(dev.name)                  
        self.CBgpu.addItem("all gpus")
        
        self.Lgpu = QLabel(self.centralWidget)
        self.Lgpu.setText('GPU settings')
        hlayout.addWidget(self.CBgpu)
        hlayout.addWidget(self.Lgpu)
        vlayout.addLayout(hlayout)     
        
        self.CBSeparateLabels = QCheckBox("Separate Labels of Stacks",self.centralWidget)
        self.CBSeparateLabels.setObjectName('Separate_Labels')
        self.CBSeparateLabels.setToolTip('Select to use different labels for each frame in an image stack')
        self.CBSeparateLabels.setEnabled(False)
        vlayout.addWidget(self.CBSeparateLabels)
                
        
        self.CBTTA = QCheckBox("Test Time Augmentation",self.centralWidget)
        self.CBTTA.setObjectName('TestTimeAugmentation')
        self.CBTTA.setToolTip('Select to use test time augmentation')
        vlayout.addWidget(self.CBTTA)
        
        self.SBBorderOverlap = LabelledSpinBox('Model Overlap',self.centralWidget)
        self.SBBorderOverlap.setObjectName('ModelOverlap')
        self.SBBorderOverlap.SpinBox.setRange(0,1024)
        self.SBBorderOverlap.setToolTip('Set tile overlap during prediction')
        vlayout.addWidget(self.SBBorderOverlap)
        

        self.CBWeightsonly = QCheckBox("Load/Save only model weights",self.centralWidget)
        self.CBWeightsonly.setObjectName('Weightsonly')
        self.CBWeightsonly.setToolTip('Select to save/load only model weights instead of complete model setup')
        vlayout.addWidget(self.CBWeightsonly)

        self.SBThreads = LabelledSpinBox('Worker threads',self.centralWidget)
        self.SBThreads.setObjectName('NumThreads')
        self.SBThreads.setToolTip('Set number of worker threads')
        self.SBThreads.SpinBox.setRange(1,multiprocessing.cpu_count())
        vlayout.addWidget(self.SBThreads)  
        
        self.SBScaleFactor = LabelledDoubleSpinBox('Image Down Scaling',self.centralWidget)
        # do not give a name as scaling is named in Training
        self.SBScaleFactor.SpinBox.setRange(0.1,1)
        self.SBScaleFactor.setToolTip('Set image reduction factor')
        self.SBScaleFactor.SpinBox.setSingleStep(0.1)
        self.SBScaleFactor.SpinBox.setDecimals(1)
        vlayout.addWidget(self.SBScaleFactor)
        

        
        
        groupBox.setLayout(vlayout)
        return groupBox
    
    def DispGroup(self):
        groupBox = QGroupBox("Display Settings")
        vlayout = QVBoxLayout(self.centralWidget)
        vlayout.setContentsMargins(1, 1, 1, 1)
        vlayout.setSpacing(1)        
        
                
        self.CBShapeNumbers = QCheckBox("Show Object Numbers",self.centralWidget)
        self.CBShapeNumbers.setToolTip('Select to show number below each shape')
        self.CBShapeNumbers.setObjectName('ShowShapeNumbers')
        vlayout.addWidget(self.CBShapeNumbers)
        
        self.CBFastDrawing = QCheckBox("Fast Drawing",self.centralWidget)
        self.CBFastDrawing.setToolTip('Reduces performance needs during drawing. Might be used for large images.')
        self.CBFastDrawing.setObjectName('FastDrawing')
        vlayout.addWidget(self.CBFastDrawing)
        

        self.SBFontSize = LabelledSpinBox('Font Size',self.centralWidget)
        self.SBFontSize.setToolTip('Set font size')
        self.SBFontSize.setObjectName('FontSize')
        self.SBFontSize.SpinBox.setRange(8,32)
        vlayout.addWidget(self.SBFontSize)

        self.SBPenSize = LabelledSpinBox('Pen Size',self.centralWidget)
        self.SBPenSize.SpinBox.setRange(1,15)
        self.SBPenSize.setObjectName('PenSize')
        self.SBPenSize.setToolTip('Set pen size')
        vlayout.addWidget(self.SBPenSize)
        
        self.h2layout = QHBoxLayout(self.centralWidget)
        self.STransparency = QSlider(Qt.Horizontal, self.centralWidget)
        self.STransparency.setToolTip('Set opacity of contours')
        self.STransparency.setObjectName('ContourTransparency')
        self.STransparency.setMinimum(0)
        self.STransparency.setMaximum(255)
        
        self.LTransparency = QLabel(self.centralWidget)
        self.LTransparency.setText(' Contour Opacity')
        self.h2layout.addWidget(self.STransparency, stretch = 1)
        self.h2layout.addWidget(self.LTransparency, stretch = 1)
        vlayout.addLayout(self.h2layout)
        groupBox.setLayout(vlayout)
        return groupBox
        

class SettingsWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(SettingsWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)
        
        self.CBShapeNumbers.setChecked(self.parent.canvas.drawShapeNumber)
        self.CBFastDrawing.setChecked(self.parent.canvas.fastPainting)
        self.CBSeparateLabels.setChecked(self.parent.separateStackLabels)
        self.SBPenSize.SpinBox.setValue(self.parent.canvas.pen_size)
        self.SBFontSize.SpinBox.setValue(self.parent.canvas.FontSize)
        self.STransparency.setValue(self.parent.canvas.ContourTransparency)
        self.SBThreads.SpinBox.setValue(self.parent.maxworker)
        self.SBScaleFactor.SpinBox.setValue(self.parent.dl.ImageScaleFactor)
        self.CBTTA.setChecked(self.parent.dl.tta)
        self.SBBorderOverlap.SpinBox.setValue(self.parent.dl.borderremoval)
        self.CBWeightsonly.setChecked(self.parent.saveload_modelweights_only)
        
        
        self.CBShapeNumbers.stateChanged.connect(self.showContourNumbers)
        self.CBFastDrawing.stateChanged.connect(self.fastDrawing)
        self.CBSeparateLabels.stateChanged.connect(self.separateLabels)
        self.SBPenSize.SpinBox.valueChanged.connect(self.setPenSize)
        self.SBFontSize.SpinBox.valueChanged.connect(self.setFontSize)
        self.STransparency.valueChanged.connect(self.setTransparency)
        self.CBgpu.currentIndexChanged.connect(self.setGPU)
        self.SBThreads.SpinBox.valueChanged.connect(self.setNumThreads)
        self.SBScaleFactor.SpinBox.valueChanged.connect(self.ScaleFactorChanged)
        self.CBTTA.stateChanged.connect(self.ttaChanged)
        self.SBBorderOverlap.SpinBox.valueChanged.connect(self.setBorder)
        self.CBWeightsonly.stateChanged.connect(self.weightsonlyChanged)
        
        self.CBgpu.setCurrentIndex(0)
        
    def setGPU(self):
        # finds cpu and gpu
        devices = tf.config.experimental.list_physical_devices()
        gpus = tf.config.experimental.list_physical_devices('GPU') 
        dev = self.CBgpu.currentIndex() - 1 
        if devices:
            try:
                if dev == -1:
                    # tensorflow decides 
                    tf.config.experimental.set_visible_devices(gpus, 'GPU')
                elif dev >= len(devices):
                    # currently unsupported
                    self.parent.PopupWarning('Use of multiple GPUs currently not supported')
                elif devices[dev].device_type == 'CPU':
                    # set cpu
                    tf.config.experimental.set_visible_devices([], 'GPU')
                else:
                    # set gpu
                    tf.config.experimental.set_visible_devices(devices[dev], 'GPU')

            except RuntimeError as e:
                # Visible devices must be set before GPUs have been initialized
                self.parent.PopupWarning('GPU settings only take effect upon program start')
    
    def fastDrawing(self):
        self.parent.canvas.fastPainting = self.CBFastDrawing.isChecked()
        
       
    def separateLabels(self):
        self.parent.separateStackLabels = self.CBSeparateLabels.isChecked()
              
    def ttaChanged(self):
        self.parent.dl.tta = self.CBTTA.isChecked()
        
    def setBorder(self):
        self.parent.dl.borderremoval = self.SBBorderOverlap.SpinBox.value()
        
    def weightsonlyChanged(self):
        self.parent.saveload_modelweights_only = self.CBWeightsonly.isChecked()
        
    def ScaleFactorChanged(self):
        self.parent.training_form.SBScaleFactor.SpinBox.setValue(self.SBScaleFactor.SpinBox.value())
        
    def silentlyUpdateScale(self, value):
        self.SBScaleFactor.SpinBox.blockSignals(True)
        self.SBScaleFactor.SpinBox.setValue(value)
        self.SBScaleFactor.SpinBox.blockSignals(False)
        
    def setNumThreads(self):
        self.parent.setWorkers(self.SBThreads.SpinBox.value())
    
    def showContourNumbers(self):
        self.parent.canvas.drawShapeNumber = self.CBShapeNumbers.isChecked()
        
    def setPenSize(self):
        self.parent.canvas.pen_size = self.SBPenSize.SpinBox.value()
        
    def setFontSize(self):
        self.parent.canvas.FontSize = self.SBFontSize.SpinBox.value()
        
    def setTransparency(self):
        self.parent.canvas.ContourTransparency = self.STransparency.value()
        
    def closeEvent(self, event):
        self.parent.canvas.checkForChanges()


import copy
import dl.models.efficientnet.model as eff
from dl.models.classification_models.models_factory import ModelsFactory

from . import inception_resnet_v2 as irv2
from . import inception_v3 as iv3


class BackbonesFactory(ModelsFactory):
    _default_feature_layers = {

        # List of layers to take features from backbone in the following order:
        # (x16, x8, x4, x2, x1) - `x4` mean that features has 4 times less spatial
        # resolution (Height x Width) than input image.

        # VGG
        'vgg16': ('block5_conv3', 'block4_conv3', 'block3_conv3', 'block2_conv2', 'block1_conv2'),
        'vgg19': ('block5_conv4', 'block4_conv4', 'block3_conv4', 'block2_conv2', 'block1_conv2'),

        # ResNets
        'resnet18': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),
        'resnet34': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),
        'resnet50': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),
        'resnet101': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),
        'resnet152': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),

        # ResNeXt
        'resnext50': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),
        'resnext101': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),

        # Inception
        'inceptionv3': (228, 86, 16, 9),
        'inceptionresnetv2': (594, 260, 16, 9),

        # DenseNet
        'densenet121': (311, 139, 51, 4),
        'densenet169': (367, 139, 51, 4),
        'densenet201': (479, 139, 51, 4),

        # SE models
        'seresnet18': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),
        'seresnet34': ('stage4_unit1_relu1', 'stage3_unit1_relu1', 'stage2_unit1_relu1', 'relu0'),
        'seresnet50': (246, 136, 62, 4),
        'seresnet101': (552, 136, 62, 4),
        'seresnet152': (858, 208, 62, 4),
        'seresnext50': (1078, 584, 254, 4),
        'seresnext101': (2472, 584, 254, 4),
        'senet154': (6884, 1625, 454, 12),

        # Mobile Nets
        'mobilenet': ('conv_pw_11_relu', 'conv_pw_5_relu', 'conv_pw_3_relu', 'conv_pw_1_relu'),
        'mobilenetv2': ('block_13_expand_relu', 'block_6_expand_relu', 'block_3_expand_relu',
                        'block_1_expand_relu'),

        # EfficientNets
        'efficientnetb0': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),
        'efficientnetb1': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),
        'efficientnetb2': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),
        'efficientnetb3': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),
        'efficientnetb4': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),
        'efficientnetb5': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),
        'efficientnetb6': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),
        'efficientnetb7': ('block6a_expand_activation', 'block4a_expand_activation',
                           'block3a_expand_activation', 'block2a_expand_activation'),

    }

    _models_update = {
        'inceptionresnetv2': [irv2.InceptionResNetV2, irv2.preprocess_input],
        'inceptionv3': [iv3.InceptionV3, iv3.preprocess_input],

        'efficientnetb0': [eff.EfficientNetB0, eff.preprocess_input],
        'efficientnetb1': [eff.EfficientNetB1, eff.preprocess_input],
        'efficientnetb2': [eff.EfficientNetB2, eff.preprocess_input],
        'efficientnetb3': [eff.EfficientNetB3, eff.preprocess_input],
        'efficientnetb4': [eff.EfficientNetB4, eff.preprocess_input],
        'efficientnetb5': [eff.EfficientNetB5, eff.preprocess_input],
        'efficientnetb6': [eff.EfficientNetB6, eff.preprocess_input],
        'efficientnetb7': [eff.EfficientNetB7, eff.preprocess_input],
    }

    # currently not supported
    _models_delete = ['resnet50v2', 'resnet101v2', 'resnet152v2',
                      'nasnetlarge', 'nasnetmobile', 'xception']

    @property
    def models(self):
        all_models = copy.copy(self._models)
        all_models.update(self._models_update)
        for k in self._models_delete:
            del all_models[k]
        return all_models

    def get_backbone(self, name, *args, **kwargs):
        model_fn, _ = self.get(name)
        model = model_fn(*args, **kwargs)
        return model

    def get_feature_layers(self, name, n=5):
        return self._default_feature_layers[name][:n]

    def get_preprocessing(self, name):
        return self.get(name)[1]


Backbones = BackbonesFactory()



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 19 11:06:36 2020

@author: Koerber
"""
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import json
from ui.ui_utils import QSetting
from ui.ui_utils import ClassList

class Settings: 
    def __init__(self,parent):
        self.parent = parent
        self.settingsDict = {}
    
    
    def saveSettings(self, path, networksettingsonly=False):
        self.collectSettings(networksettingsonly)
        try:
            with open(path, 'w', encoding='utf-8') as f:
                json.dump(self.settingsDict, f, ensure_ascii=False, indent=4)
        except:
            # can not save
            pass

        
    def loadSettings(self, path):
        self.settingsDict.clear()
        try:
            with open(path, 'r', encoding='utf-8') as f:
                self.settingsDict = json.load(f)
            self.applySettings()
            self.parent.settingsLoaded()
        except:
            # can not load
            pass

            
    
    def collectSettings(self, nnsettingsonly):
        self.settingsDict.clear()
        self.addObjectSetting(QCheckBox, QCheckBox.isChecked, nnsettingsonly)
        self.addObjectSetting(QRadioButton, QRadioButton.isChecked, nnsettingsonly)
        self.addObjectSetting(QSlider, QSlider.value, nnsettingsonly)
        self.addObjectSetting(QComboBox, QComboBox.currentIndex, nnsettingsonly)
        self.addObjectSetting(QDoubleSpinBox, QDoubleSpinBox.value, nnsettingsonly)
        self.addObjectSetting(QSpinBox, QSpinBox.value, nnsettingsonly)
        self.addObjectSetting(QGroupBox, QGroupBox.isChecked, nnsettingsonly)
        self.addObjectSetting(QLineEdit, QLineEdit.text, nnsettingsonly)
        self.addObjectSetting(QSetting, QSetting.value, nnsettingsonly)

   
    def addObjectSetting(self, objectType, function, nnsettingsonly):
        items = self.parent.findChildren(objectType)
        for i in items:
            if i.objectName() != '' and i.objectName() != 'qt_spinbox_lineedit': # spinbox line edit has qt_spinbox_lineedit as default name
                if not nnsettingsonly:
                    self.settingsDict[i.objectName()] = function(i)
                else:
                    if i.objectName().startswith('nn'):
                        self.settingsDict[i.objectName()] = function(i)
        
            
    def applySettings(self):
        self.parent.setNumberOfClasses(self.settingsDict['nn_NumOfClasses'])
        for obj in self.settingsDict:
            widget = self.parent.findChild(QWidget, obj)
            if isinstance(widget, QCheckBox):
                widget.setChecked(self.settingsDict[obj])
            elif isinstance(widget, QRadioButton):
                widget.setChecked(self.settingsDict[obj])
            elif isinstance(widget, QSlider):
                widget.setValue(self.settingsDict[obj])
            elif isinstance(widget, QComboBox):
                widget.setCurrentIndex(self.settingsDict[obj])
            elif isinstance(widget, QDoubleSpinBox):
                widget.setValue(self.settingsDict[obj])
            elif isinstance(widget, QSpinBox):
                widget.setValue(self.settingsDict[obj])
            elif isinstance(widget, QGroupBox):
                widget.setChecked(self.settingsDict[obj])  
            elif isinstance(widget, QLineEdit):
                widget.setText(self.settingsDict[obj])  
            elif isinstance(widget, QSetting):
                widget.setValue(self.settingsDict[obj])  
            # else:
                # do nothing

        



# this is a clone of https://github.com/keras-team/keras-applications
# with changes mostly regarding model weight downloads
# licence MIT, https://github.com/keras-team/keras-applications/blob/master/LICENSE


"""Enables dynamic setting of underlying Keras module.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

_KERAS_BACKEND = None
_KERAS_LAYERS = None
_KERAS_MODELS = None
_KERAS_UTILS = None


def get_submodules_from_kwargs(kwargs):
    backend = kwargs.get('backend', _KERAS_BACKEND)
    layers = kwargs.get('layers', _KERAS_LAYERS)
    models = kwargs.get('models', _KERAS_MODELS)
    utils = kwargs.get('utils', _KERAS_UTILS)
    for key in kwargs.keys():
        if key not in ['backend', 'layers', 'models', 'utils']:
            raise TypeError('Invalid keyword argument: %s', key)
    return backend, layers, models, utils


def correct_pad(backend, inputs, kernel_size):
    """Returns a tuple for zero-padding for 2D convolution with downsampling.

    # Arguments
        input_size: An integer or tuple/list of 2 integers.
        kernel_size: An integer or tuple/list of 2 integers.

    # Returns
        A tuple.
    """
    img_dim = 2 if backend.image_data_format() == 'channels_first' else 1
    input_size = backend.int_shape(inputs)[img_dim:(img_dim + 2)]

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)

    if input_size[0] is None:
        adjust = (1, 1)
    else:
        adjust = (1 - input_size[0] % 2, 1 - input_size[1] % 2)

    correct = (kernel_size[0] // 2, kernel_size[1] // 2)

    return ((correct[0] - adjust[0], correct[0]),
            (correct[1] - adjust[1], correct[1]))

__version__ = '1.0.8'


from . import vgg16
from . import vgg19
from . import resnet50
from . import inception_v3
from . import inception_resnet_v2
from . import xception
from . import mobilenet
from . import mobilenet_v2
from . import densenet
from . import nasnet
from . import resnet
from . import resnet_v2
from . import resnext



# this is a clone from https://github.com/qubvel/classification_models
# with minor changes
# licence MIT, https://github.com/qubvel/classification_models/blob/master/LICENSE

import dl.models.keras_applications as ka

def get_submodules_from_kwargs(kwargs):
    backend = kwargs.get('backend', ka._KERAS_BACKEND)
    layers = kwargs.get('layers', ka._KERAS_LAYERS)
    models = kwargs.get('models', ka._KERAS_MODELS)
    utils = kwargs.get('utils', ka._KERAS_UTILS)
    return backend, layers, models, utils


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 13:43:00 2020

@author: Koerber
"""


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.style import styleForm


class Window(object):
    def setupUi(self, Form):

        Form.setWindowTitle('Title') 
        styleForm(Form)
        
        self.centralWidget = QWidget(Form)

        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.vlayout.setSpacing(2)
        self.centralWidget.setLayout(self.vlayout)

        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())

class TitleWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(TitleWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)
        


from dl.models.keras_applications import get_submodules_from_kwargs


def freeze_model(model, **kwargs):
    """Set all layers non trainable, excluding BatchNormalization layers"""
    _, layers, _, _ = get_submodules_from_kwargs(kwargs)
    for layer in model.layers:
        if not isinstance(layer, layers.BatchNormalization):
            layer.trainable = False
    return



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 15:22:17 2020

@author: Koerber
"""


# PointContour_ID = 987654321
import numpy as np
import math
from utils.shapes.Shape import Shape, Shapes, PointContour_ID
from utils.shapes.Contour import findContours, packContours, unpackContours, drawbackgroundToLabel
import cv2


class Points(Shapes):
    def __init__(self):
        super(Points,self).__init__(PointContour_ID)
        self.mindistance = 3

    def addShapes(self,pts):
        # atm it is not checked whether pts are within self.mindistance
        ext = [x for x in pts if x.labeltype == self.labeltype and not self.ShapeAlreadyExist(x)]
        self.shapes.extend(ext)

    def ShapeAlreadyExist(self,s):
        return next((True for elem in self.shapes if distance_p1_p2(elem.coordinates,s.coordinates) < self.mindistance), False)
    
    def load(self, filename):
        self.shapes, _ = loadPoints(filename)
        
    def save(self, filename):
        savePoints(self.shapes, filename)
        
    def deleteShapes(self, points):
        for p in points:
            self.deleteShape(p)
        
    def getShapeOfClass_x(self,x, point, distance):
        if self.empty():
            return None
        point = point[0]
        shapes = self.getShapesOfClass_x(x)
        if not shapes:
            return None
        dist_list = [distance_p1_p2(p.coordinates,point) for p in shapes]
        target = shapes[dist_list.index(min(dist_list))]
        if distance_p1_p2(target.coordinates,point) > distance:
            return None
        return target
      
    def getShape(self, point, distance):
        if self.empty():
            return None
        point = point[0]
        dist_list = [distance_p1_p2(p.coordinates,point) for p in self.shapes]
        target = self.shapes[dist_list.index(min(dist_list))]
        if distance_p1_p2(target.coordinates,point) > distance:
            return None
        return target
    

class Point(Shape):
    def __init__(self, classlabel, point):
        self.coordinates = np.asarray(point).astype(int)
        super(Point,self).__init__(classlabel, PointContour_ID)
        
    def getPosition(self):
        return (self.coordinates[0], self.coordinates[1])

    def inside(self, point, distance):
        if distance_p1_p2(self.coordinates,point) <= distance:
            return True
        else:
            return False
            

# utilities
def distance_p1_p2(p1,p2):
    return math.sqrt((p1[0]-p2[0])**2 + (p1[1]-p2[1])**2)

def savePoints(points, filename, background = None):
    bg = packContours(background)
    pts = []
    if len(points) > 0:   
        f1 = lambda x: np.array((x.classlabel, x.objectNumber),dtype=int)
        f2 = lambda x: x.coordinates
        labels = [f1(x) for x in points]
        coords = [f2(x) for x in points]
        pts = [points[0].labeltype, np.stack(labels), np.stack(coords)]
    else:
        if bg:
            pts = [PointContour_ID, [], []]
    if bg:
        pts.extend(bg)
    if pts:
        np.savez(filename, *pts)

    
def loadPoints(filename):
    container = np.load(filename)
    data = [container[key] for key in container] 
    ret = []
    bg = None
    if data[0] == PointContour_ID:
        label = data[1]
        array = data[2]
        ret = [Point(x,y) for x,y in zip(label,array)]
        
        if len(data) > 3:
            bg = unpackContours(data[3:])
            
    return ret, bg

def extractPointsFromLabel(image, offset):
    numclasses = np.max(image)
    points = []
    for i in range(1,numclasses+1):
        pts = np.where(image == i)
        p = zip(pts[1], pts[0])
        points.extend([Point(i, (x+offset[0],y+offset[1])) for x,y in p])

    return points


def extractPointsFromContours(image, minsize, offset = (0,0)):
    contours, _ = findContours(image, ext_only = True, offset=offset)
    points = []
    for c in contours:
        c = cv2.convexHull(c, False)
        if cv2.contourArea(c) < minsize:
            continue
        M = cv2.moments(c)
        if M["m00"] == 0:
            continue
        cX = int(M["m10"] / M["m00"])
        cY = int(M["m01"] / M["m00"])
        points.append(Point(1, (cX, cY)))
    return points

def drawPointsToLabel(image, points, bg_contours): 
    image = drawbackgroundToLabel(image, bg_contours)
    
    for p in points:
        image[p.coordinates[1],p.coordinates[0]] = p.classlabel      
    return image



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 10 11:28:44 2019

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

from ui.ui_utils import ClassList, ClassWidget, DCDButton, LabelledSpinBox
from ui.style import styleForm, getHighlightColor, getBackgroundColor, getBrightColor

class MainWindow(object):
    def setupUi(self, Form):
        screen_resolution = QDesktopWidget().screenGeometry()


        height = int(0.85*screen_resolution.height())
        width = int(height * 4/3)

        Form.setWindowTitle('Microscopic Image Analyzer') 
        Form.setWindowIcon(QIcon('icons/logo.png'))
        Form.resize(width, height)
        Form.setStyleSheet("background: " + getBackgroundColor(asString = True))
        styleForm(Form)

        self.centralWidget = QWidget(Form)

        sizePolicy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.centralWidget.sizePolicy().hasHeightForWidth())
        self.centralWidget.setSizePolicy(sizePolicy)
        
        ###### menu
        self.menubar = QMenuBar(self.centralWidget)
        
        self.FileMenu = self.menubar.addMenu("File")
        self.ASetTrainingFolder = QAction("Training Folder", self.centralWidget)
        self.FileMenu.addAction(self.ASetTrainingFolder)
        
        self.ASetTestFolder = QAction("Test Folder", self.centralWidget)
        self.FileMenu.addAction(self.ASetTestFolder)
        
        self.ASaveSettings = QAction("Save Settings", self.centralWidget)
        self.FileMenu.addAction(self.ASaveSettings)
        self.ALoadSettings = QAction("Load Settings", self.centralWidget)
        self.FileMenu.addAction(self.ALoadSettings)
        
        self.ASaveDeepLearning = QAction("Save DL Object", self.centralWidget)
        self.FileMenu.addAction(self.ASaveDeepLearning)
        
        
        self.FileMenu.addSeparator()
        self.AExit = QAction("Quit", self)
        self.FileMenu.addAction(self.AExit)
        self.EditMenu = self.menubar.addMenu("Edit")
        self.ASettings = QAction("Settings")
        self.EditMenu.addAction(self.ASettings)
        self.AboutMenu = self.menubar.addMenu("About")
        self.setMenuBar(self.menubar)
        self.hlayout = QHBoxLayout(self.centralWidget)
        self.hlayout.setContentsMargins(3, 3, 3, 3)
        self.hlayout.setSpacing(6)
        self.centralWidget.setLayout(self.hlayout)
        
        self.vlayout = QVBoxLayout()
        verticalSpacer = QSpacerItem(20, 40)
        self.vlayout.addItem(verticalSpacer)
        
        
        ## vertical layout - buttons
        self.BSettings = DCDButton(self.centralWidget, 'Settings')
        self.BSettings.setToolTip('Open settings sindow')
        self.BSettings.setIcon(QIcon('icons/settings.png'))
        self.BSettings.setFlat(True)

        self.vlayout.addWidget(self.BSettings)
        
        self.CBLearningMode = QComboBox(self.centralWidget)
        self.CBLearningMode.addItem("Classification")
        self.CBLearningMode.addItem("Object Detection")
        self.CBLearningMode.addItem("Segmentation")
        self.CBLearningMode.setToolTip('Select analysis mode')        
        self.CBLearningMode.setObjectName("nn_Learning_Mode")
        self.CBLearningMode.setEditable(False)
        self.CBLearningMode.setFocus(False)
        self.vlayout.addWidget(self.CBLearningMode)
        
        bwidth = 28
        bheight = 28
        
        self.TrainImagelayout = QHBoxLayout(self.centralWidget)
        self.BLoadTrainImages = DCDButton(self.centralWidget)
        self.BLoadTrainImages.setToolTip('Set folder with training images')
        self.BLoadTrainImages.setMaximumSize(bwidth, bheight)
        self.BLoadTrainImages.setIcon(QIcon('icons/load.png'))
         

        self.BTrainImageFolder = DCDButton(self.centralWidget, "Set Training Image Folder")
        self.BTrainImageFolder.setToolTip('Training image folder')
        self.BTrainImageFolder.setEnabled(False)
        self.TrainImagelayout.addWidget(self.BLoadTrainImages)
        self.TrainImagelayout.addWidget(self.BTrainImageFolder)
        self.vlayout.addItem(self.TrainImagelayout)
        
        self.TestImagelayout = QHBoxLayout(self.centralWidget)
        self.BLoadTestImages = DCDButton(self.centralWidget)
        self.BLoadTestImages.setMaximumSize(bwidth, bheight)
        self.BLoadTestImages.setToolTip('Set folder with prediction images')
        self.BLoadTestImages.setIcon(QIcon('icons/load.png'))

        self.BTestImageFolder = DCDButton(self.centralWidget,"Set Test Image Folder")
        self.BTestImageFolder.setToolTip('Prediction image folder')
        self.BTestImageFolder.setEnabled(False)
        self.TestImagelayout.addWidget(self.BLoadTestImages)
        self.TestImagelayout.addWidget(self.BTestImageFolder)
        self.vlayout.addItem(self.TestImagelayout)

        self.TVFiles = QTreeView(self.centralWidget)
        self.TVFiles.setIndentation(8)
        self.TVFiles.setVisible(False)
        self.vlayout.addWidget(self.TVFiles)
        
        line = QFrame(self.centralWidget)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setStyleSheet("background: " + getBrightColor(asString = True))
        self.vlayout.addWidget(line)
        
        
        # settings for segmentation
        self.SegmentationSettingslayout = QHBoxLayout(self.centralWidget)
        self.SegmentationSettingslayout.setContentsMargins(0, 0, 0, 0)
        layout0 = QVBoxLayout(self.centralWidget)
        self.Bautoseg = DCDButton(self.centralWidget)
        self.Bautoseg.setToolTip('press to perform auto segmentation')
        self.Bautoseg.setText('Auto Seg')
        self.Bautoseg.setIcon(QIcon('icons/magicwand.png'))
        
        self.Bassist = DCDButton(self.centralWidget)
        self.Bassist.setText('DEXTR')
        self.Bassist.setObjectName('assist')
        self.Bassist.setToolTip('Select assisted extreme point segmentation. Mark extreme points (top, left, right, bottom) of target object.')
        self.Bassist.setIcon(QIcon('icons/dextr.png'))
        layout0.addWidget(self.Bautoseg)
        layout0.addWidget(self.Bassist)

        layout = QVBoxLayout(self.centralWidget)
        self.CBSmartMode = QCheckBox("Smart Mode",self.centralWidget)
        self.CBSmartMode.setToolTip("Check to activate smart segmentation mode")
        self.CBSmartMode.setObjectName('SmartMode')
        self.CBInner = QCheckBox("Inner Contours",self.centralWidget)
        self.CBInner.setToolTip("Check to allow contours to have holes")
        self.CBInner.setObjectName('InnerContours')
        self.CBInner.setChecked(True)
        layout.addWidget(self.CBSmartMode)
        layout.addWidget(self.CBInner)

        self.SegmentationSettingslayout.addLayout(layout)
        self.SegmentationSettingslayout.addLayout(layout0)
        self.SegmentationSettings = QFrame()
        self.SegmentationSettings.setLayout(self.SegmentationSettingslayout)
        self.SegmentationSettings.hide()
        self.vlayout.addWidget(self.SegmentationSettings)
        
        
        # settings for object detection
        self.DetectionSettingslayout = QHBoxLayout(self.centralWidget)
        self.DetectionSettingslayout.setContentsMargins(0, 0, 0, 0)
        self.Bautodect = DCDButton(self.centralWidget)
        self.Bautodect.setToolTip('press to perform auto detection')
        self.Bautodect.setText('Auto Detection')
        self.Bautodect.setIcon(QIcon('icons/magicwand.png'))
        
        self.DetectionSettingslayout.addWidget(self.Bautodect)
        self.DetectionSettings = QFrame()
        self.DetectionSettings.setLayout(self.DetectionSettingslayout)
        self.DetectionSettings.hide()
        self.vlayout.addWidget(self.DetectionSettings)
        

        
        # tools:
        # drag tool
        self.Bdrag = DCDButton(self.centralWidget)
        self.Bdrag.setToolTip('Select drag tool')
        self.Bdrag.setVisible(False)
        self.Bdrag.setObjectName('drag')
        self.Bdrag.setMaximumSize(bwidth, bheight)
        self.Bdrag.setIcon(QIcon('icons/drag.png'))
        #draw tool
        self.Bdraw = DCDButton(self.centralWidget)
        self.Bdraw.setMaximumSize(bwidth, bheight)
        self.Bdraw.setVisible(False)
        self.Bdraw.setObjectName('draw')
        self.Bdraw.setToolTip('Select drawing tool')
        self.Bdraw.setIcon(QIcon('icons/draw.png'))
        # settings for drawing tool
        self.DrawSettingslayout = QHBoxLayout(self.centralWidget)
        self.DrawSettingslayout.setContentsMargins(0, 0, 0, 0)
        self.CBAddShape = QRadioButton(self.centralWidget)
        self.CBAddShape.setText("Add ")
        self.CBAddShape.setChecked(True)
        self.CBAddShape.setToolTip('Toggle to add shape')
        self.CBDelShape = QRadioButton(self.centralWidget)
        self.CBDelShape.setText("Delete  ")
        self.CBDelShape.setToolTip('Toggle to delete shape')
        self.CBSliceShape = QRadioButton(self.centralWidget)
        self.CBSliceShape.setText("Slice  ")
        self.CBSliceShape.setToolTip('Toggle to slice shapes')
        self.DrawSettingslayout.addWidget(self.CBAddShape)
        self.DrawSettingslayout.addWidget(self.CBDelShape)
        self.DrawSettingslayout.addWidget(self.CBSliceShape)
        self.DrawSettings = QFrame()
        self.DrawSettings.setLayout(self.DrawSettingslayout)
        self.DrawSettings.hide()
        # polyline tool
        self.Bpoly = DCDButton(self.centralWidget)
        self.Bpoly.setMaximumSize(bwidth, bheight)
        self.Bpoly.setVisible(False)
        self.Bpoly.setObjectName('poly')
        self.Bpoly.setToolTip('Select polygon tool')
        self.Bpoly.setIcon(QIcon('icons/poly.png'))
        # assign tool
        self.Bassign = DCDButton(self.centralWidget)
        self.Bassign.setMaximumSize(bwidth, bheight)
        self.Bassign.setVisible(False)
        self.Bassign.setObjectName('assign')
        self.Bassign.setToolTip('Select assign class tool')
        self.Bassign.setIcon(QIcon('icons/assign.png'))
        # extend tool
        self.Bextend = DCDButton(self.centralWidget)
        self.Bextend.setMaximumSize(bwidth, bheight)
        self.Bextend.setVisible(False)
        self.Bextend.setObjectName('extend')
        self.Bextend.setToolTip('Select extend/erase tool')
        self.Bextend.setIcon(QIcon('icons/expand.png'))
        # settings for extend tool
        self.ExtendSettingslayout = QHBoxLayout(self.centralWidget)
        self.ExtendSettingslayout.setContentsMargins(0, 0, 0, 0)
        self.CBExtend = QRadioButton(self.centralWidget)
        self.CBExtend.setText("Expand ")
        self.CBExtend.setChecked(True)
        self.CBExtend.setToolTip('Toggle to extend tool')
        self.CBErase = QRadioButton(self.centralWidget)
        self.CBErase.setText("Erase  ")
        self.CBErase.setToolTip('Toggle  to erase tool')
        self.SSize = QSlider(Qt.Horizontal, self.centralWidget)
        self.SSize.setMinimum(1)
        self.SSize.setMaximum(40)
        self.SSize.setValue(10)
        self.SSize.setToolTip('Set marker size')
        self.SSize.setMaximumWidth((self.CBExtend.width()+self.CBErase.width())/4)
        self.ExtendSettingslayout.addWidget(self.CBExtend)
        self.ExtendSettingslayout.addWidget(self.CBErase)
        self.ExtendSettingslayout.addWidget(self.SSize)
        self.ExtendSettings = QFrame()
        self.ExtendSettings.setLayout(self.ExtendSettingslayout)
        self.ExtendSettings.hide()
        # delete tool
        self.Bdelete = DCDButton(self.centralWidget)
        self.Bdelete.setMaximumSize(bwidth, bheight)
        self.Bdelete.setVisible(False)
        self.Bdelete.setObjectName('delete')
        self.Bdelete.setToolTip('Select delete tool')
        self.Bdelete.setIcon(QIcon('icons/delete.png'))
        # set object tool
        self.BsetObject = DCDButton(self.centralWidget)
        self.BsetObject.setObjectName('point')
        self.BsetObject.setVisible(False)
        self.BsetObject.setMaximumSize(bwidth, bheight)
        self.BsetObject.setToolTip('Select set object position tool')
        self.BsetObject.setIcon(QIcon('icons/addobject.png'))       
        # shift object tool
        self.BshiftObject = DCDButton(self.centralWidget)
        self.BshiftObject.setMaximumSize(bwidth, bheight)
        self.BshiftObject.setVisible(False)
        self.BshiftObject.setObjectName('shift')
        self.BshiftObject.setToolTip('Select shift object tool')
        self.BshiftObject.setIcon(QIcon('icons/shift.png'))
        # assign class tool
        self.BassignClass = DCDButton(self.centralWidget)
        self.BassignClass.setMaximumSize(bwidth, bheight)
        self.BassignClass.setObjectName('assignimageclass')
        self.BassignClass.setToolTip('Assign current class')
        self.BassignClass.setIcon(QIcon('icons/assignclass.png'))
        # set class tool
        self.BsetClass = DCDButton(self.centralWidget)
        self.BsetClass.setMaximumSize(bwidth, bheight)
        self.BsetClass.setObjectName('setimageclass')
        self.BsetClass.setToolTip('Assign current class and continue to next image')
        self.BsetClass.setIcon(QIcon('icons/setclass.png'))
        
        # tool buttons
        self.ToolButtonslayout = QHBoxLayout(self.centralWidget)
        self.ToolButtonslayout.setContentsMargins(0, 0, 0, 0)
        self.ToolButtons = QFrame()
        self.ToolButtons.setMaximumWidth(self.CBExtend.width()*3)
        self.ToolButtons.setLayout(self.ToolButtonslayout)
        self.ToolButtonslayout.addWidget(self.Bdrag)
        self.ToolButtonslayout.addWidget(self.Bdraw)
        self.ToolButtonslayout.addWidget(self.BsetObject)
        self.ToolButtonslayout.addWidget(self.Bpoly)
        self.ToolButtonslayout.addWidget(self.BshiftObject)
        self.ToolButtonslayout.addWidget(self.Bassign)
        self.ToolButtonslayout.addWidget(self.Bextend)
        self.ToolButtonslayout.addWidget(self.Bdelete)

        self.ClassificationToolButtonslayout = QHBoxLayout(self.centralWidget)
        self.ClassificationToolButtonslayout.setContentsMargins(0, 0, 0, 0)
        self.ClassificationToolButtons = QFrame()
        self.ClassificationToolButtons.setLayout(self.ClassificationToolButtonslayout)
        self.ClassificationToolButtons.hide()
        self.ClassificationToolButtonslayout.addWidget(self.BassignClass)
        self.ClassificationToolButtonslayout.addWidget(self.BsetClass)
        
        self.vlayout.addWidget(self.ClassificationToolButtons)
        self.vlayout.addWidget(self.ToolButtons)

        
        # add settings frames
        self.vlayout.addWidget(self.DrawSettings)
        self.vlayout.addWidget(self.ExtendSettings)

        
        
        line = QFrame(self.centralWidget)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setStyleSheet("background: " + getBrightColor(asString = True))
        self.vlayout.addWidget(line)

        # tracking mode
        self.Trackinglayout = QVBoxLayout(self.centralWidget)
        self.Trackinglayout.setContentsMargins(0, 0, 0, 0)
        self.Tracking = QFrame()
        self.Tracking.setLayout(self.Trackinglayout)

        self.BTrack = DCDButton(self.centralWidget, 'Calculate Tracking')
        self.BTrack.setToolTip('Press to calculate tracking')
        self.BTrack.setIcon(QIcon('icons/tracking.png'))

        self.trcolorlayout = QHBoxLayout(self.centralWidget)
        self.CBtrackcolor = QCheckBox("Color Coding",self.centralWidget)
        self.CBtrackcolor.setToolTip("Check to change coloring by objects, uncheck to color by classes")
        self.CBtrackcolor.setObjectName('TrackColor')
        self.CBtrackcolor.setChecked(True)

        self.CBshowtrack = QCheckBox("Show Track",self.centralWidget)
        self.CBshowtrack.setToolTip("Check to show track")
        self.CBshowtrack.setObjectName('ShowTrack')
        self.CBshowtrack.setChecked(True)
        self.trcolorlayout.addWidget(self.CBtrackcolor)
        self.trcolorlayout.addWidget(self.CBshowtrack)

        self.objNumLayout = QHBoxLayout(self.centralWidget)
        self.BSetObjectNumber = DCDButton(self.centralWidget)
        self.BSetObjectNumber.setMaximumSize(bwidth, bheight)
        self.BSetObjectNumber.setObjectName('objectnumber')
        self.BSetObjectNumber.setIcon(QIcon('icons/objectnumber.png'))
        self.BSetObjectNumber.setToolTip('Select set object number tool')

        self.BDeleteObject = DCDButton(self.centralWidget)
        self.BDeleteObject.setMaximumSize(bwidth, bheight)
        self.BDeleteObject.setObjectName('deleteobject')
        self.BDeleteObject.setIcon(QIcon('icons/delete.png'))
        self.BDeleteObject.setToolTip('Select delete object tool')

        self.BChangeObjectColor = DCDButton(self.centralWidget)
        self.BChangeObjectColor.setMaximumSize(bwidth, bheight)
        self.BChangeObjectColor.setObjectName('objectcolor')
        self.BChangeObjectColor.setIcon(QIcon('icons/objectcolor.png'))
        self.BChangeObjectColor.setToolTip('Select change color tool')

        self.objNumLayout.addWidget(self.BSetObjectNumber)
        self.objNumLayout.addWidget(self.BChangeObjectColor)
        self.objNumLayout.addWidget(self.BDeleteObject)

        self.Trackinglayout.addWidget(self.BTrack)
        self.Trackinglayout.addLayout(self.trcolorlayout)
        self.Trackinglayout.addLayout(self.objNumLayout)

        line = QFrame(self.centralWidget)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setStyleSheet("background: " + getBrightColor(asString = True))
        self.Trackinglayout.addWidget(line)

        self.Tracking.setVisible(False)
        self.vlayout.addWidget(self.Tracking)


        
        self.ralayout = QHBoxLayout(self.centralWidget)
        self.Bundo = DCDButton(self.centralWidget, "Undo")
        self.Bundo.setIcon(QIcon('icons/reset.png'))
        self.Bundo.setToolTip('Undo drawing') 
        self.Bclear = DCDButton(self.centralWidget, "Clear Image")
        self.Bclear.setIcon(QIcon('icons/clear.png'))
        self.Bclear.setToolTip('Reset image')  
        self.ralayout.addWidget(self.Bundo) 
        self.ralayout.addWidget(self.Bclear)    
        self.vlayout.addItem(self.ralayout) 
        
        line = QFrame(self.centralWidget)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setStyleSheet("background: " + getBrightColor(asString = True))
        self.vlayout.addWidget(line)

        self.trlayout = QHBoxLayout(self.centralWidget)
        self.Btrain = DCDButton(self.centralWidget, "Train Model")
        self.Btrain.setToolTip('Open training window')
        self.Btrain.setIcon(QIcon('icons/train.png'))

                
        self.Bresetmodel = DCDButton(self.centralWidget, "Reset Model")
        self.Bresetmodel.setIcon(QIcon('icons/clear.png'))
        self.Bresetmodel.setToolTip('reset trained model')

        self.trlayout.addWidget(self.Btrain)    
        self.trlayout.addWidget(self.Bresetmodel)    
        self.vlayout.addItem(self.trlayout) 
            

        
        self.predictlayout = QHBoxLayout(self.centralWidget)
        self.Bpredictall = DCDButton(self.centralWidget, "Predict All")
        self.Bpredictall.setIcon(QIcon('icons/predictall.png'))
        self.Bpredictall.setToolTip('Predict all images in prediction folder')
        self.Bpredict = DCDButton(self.centralWidget, "Predict Current")
        self.Bpredict.setIcon(QIcon('icons/predict.png'))
        self.Bpredict.setToolTip('Predict current field of view')
        self.predictlayout.addWidget(self.Bpredictall)    
        self.predictlayout.addWidget(self.Bpredict)
        
        self.vlayout.addItem(self.predictlayout) 
        
        self.lslayout = QHBoxLayout(self.centralWidget)
        self.Bloadmodel = DCDButton(self.centralWidget, "Load Model")
        self.Bloadmodel.setIcon(QIcon('icons/loadmodel.png'))
        self.Bloadmodel.setToolTip('Load trained model')
        
        self.Bsavemodel = DCDButton(self.centralWidget, "Save Model")
        self.Bsavemodel.setIcon(QIcon('icons/savemodel.png'))
        self.Bsavemodel.setToolTip('Save current model')
        self.lslayout.addWidget(self.Bloadmodel)    
        self.lslayout.addWidget(self.Bsavemodel)
        
        self.vlayout.addItem(self.lslayout) 
        

        
        ##############
        self.classList = ClassList(self)
        self.classList.addClass('background')
        self.classList.addClass()
        self.vlayout.addWidget(self.classList)
        
        
        
        

        
        self.add_del_layout = QHBoxLayout(self.centralWidget)
        self.Baddclass = DCDButton(self.centralWidget)
        self.Baddclass.setIcon(QIcon('icons/addobject.png'))
        self.Baddclass.setToolTip('Add a new class')
        self.Baddclass.setMaximumHeight(bheight)
        self.Baddclass.setStyleSheet('text-align:center')
        self.Baddclass.resetStyle()      
        self.add_del_layout.addWidget(self.Baddclass)
        self.Bdelclass = DCDButton(self.centralWidget)
        self.Bdelclass.setIcon(QIcon('icons/delete.png'))
        self.Bdelclass.setToolTip('Delete class')
        self.Bdelclass.setMaximumHeight(bheight)
        self.Bdelclass.setStyleSheet('text-align:center')
        self.Bdelclass.resetStyle()  
        self.add_del_layout.addWidget(self.Bdelclass)
        self.vlayout.addItem(self.add_del_layout)
        line = QFrame(self.centralWidget)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setStyleSheet("background: " + getBrightColor(asString = True))
        self.vlayout.addWidget(line)
        
        self.BPostProcessing = DCDButton(self.centralWidget, "Postprocessing")
        self.BPostProcessing.setObjectName("BPostProcessing")
        self.BPostProcessing.setIcon(QIcon('icons/postprocessing.png'))
        self.BPostProcessing.setToolTip('Open postprocessing window')
        self.vlayout.addWidget(self.BPostProcessing)
        
        self.Bresults = DCDButton(self.centralWidget, "Results")
        self.Bresults.setIcon(QIcon('icons/results.png'))
        self.Bresults.setToolTip('Open results window')
        self.vlayout.addWidget(self.Bresults)
        
        verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.vlayout.addItem(verticalSpacer)
        
        self.CBKeep = QCheckBox("Keep Settings",self.centralWidget)
        self.CBKeep.setToolTip("Check to keep brightness/contrast and fov settings (recommended for time series)")
        self.CBKeep.setObjectName('KeepImageSettings')
        self.CBKeep.setChecked(False)
        self.vlayout.addWidget(self.CBKeep)

        brightnesslayout = QHBoxLayout(self.centralWidget)
        self.SBrightness = QSlider(Qt.Horizontal, self.centralWidget)
        self.SBrightness.setMinimum(-200)
        self.SBrightness.setMaximum(200)
        self.SBrightness.setValue(0)
        self.SBrightness.setToolTip('Change Image Brightness')
        self.LBrightness = QLabel(self.centralWidget)
        self.LBrightness.setText(' Brightness')
        brightnesslayout.addWidget(self.SBrightness, stretch = 2)
        brightnesslayout.addWidget(self.LBrightness, stretch = 1)
        
        contrastlayout = QHBoxLayout(self.centralWidget)
        self.SContrast = QSlider(Qt.Horizontal, self.centralWidget)
        self.SContrast.setMinimum(-50)
        self.SContrast.setMaximum(250)
        self.SContrast.setValue(0)
        self.SContrast.setToolTip('Change Image Contrast')
        self.LContrast = QLabel(self.centralWidget)
        self.LContrast.setText(' Contrast')
        contrastlayout.addWidget(self.SContrast, stretch = 2)
        contrastlayout.addWidget(self.LContrast, stretch = 1)
        
        self.vlayout.addItem(brightnesslayout)
        self.vlayout.addItem(contrastlayout)
        line = QFrame(self.centralWidget)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setStyleSheet("background: " + getBrightColor(asString = True))
        self.vlayout.addWidget(line)
        
        self.nextprev_layout = QHBoxLayout(self.centralWidget)
        self.Bprev = DCDButton(self.centralWidget)
        self.Bprev.setIcon(QIcon('icons/previous.png'))
        self.Bprev.setStyleSheet('text-align:center')
        self.Bprev.setToolTip('Previous image')
        self.Bprev.resetStyle()   
        self.nextprev_layout.addWidget(self.Bprev)
        self.Bnext = DCDButton(self.centralWidget)
        self.Bnext.setIcon(QIcon('icons/next.png'))
        self.Bnext.setStyleSheet('text-align:center')
        self.Bnext.setToolTip('Next image')
        self.Bnext.resetStyle()  
        self.nextprev_layout.addWidget(self.Bnext)
        self.vlayout.addItem(self.nextprev_layout)
        
       
        ## horizontal layout - buttons + canvas    
        self.hlayout.addLayout(self.vlayout)
        horizontalspacer = QSpacerItem(20, 40, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.hlayout.addItem(horizontalspacer)
        self.canvas = QLabel(self.centralWidget)
        self.canvas.setText("")
        self.canvas.setObjectName("canvas")
        self.canvas.setPixmap(QPixmap(width, height))
        self.hlayout.addWidget(self.canvas)
        self.SFrame = QSlider(Qt.Vertical, self.centralWidget)
        self.SFrame.setMinimum(1)
        self.SFrame.setMaximum(10)
        self.SFrame.setValue(1)
        self.SFrame.setToolTip('Change Frame of Stack')
        self.SFrame.hide()
        self.hlayout.addWidget(self.SFrame)
        
        horizontalspacer = QSpacerItem(20, 40, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.hlayout.addItem(horizontalspacer)
        
        
        self.StatusFile = QLabel(self.centralWidget)
        self.StatusFile.setText('No Image Displayed')
        self.StatusFile.setObjectName("StatusFile")
        self.statusBar().addPermanentWidget(self.StatusFile, stretch = 1)
        Form.setCentralWidget(self.centralWidget)
        
        self.StatusProgress = QProgressBar(self.centralWidget)
        self.StatusProgress.setMaximumWidth(350)
        self.StatusProgress.hide()
        self.statusBar().addPermanentWidget(self.StatusProgress, stretch = 1)
        
        self.Status = QLabel(self.centralWidget)
        self.Status.setText('Initialized')
        self.Status.setObjectName("Status")
        self.statusBar().addPermanentWidget(self.Status, stretch = 0)
        

 

        
        
        